from asyncio import create_subprocess_exec, subprocess
import logging
import os
from pathlib import Path
import tempfile
from typing import (
    Any, Optional,
    Iterable,
    Mapping, MutableMapping,
)

from .logging import BraceStyleAdapter

logger = BraceStyleAdapter(logging.getLogger())


async def write_file(
    variables: Mapping[str, Any],
    filename: str,
    body: Iterable[str],
    mode: str = '644',
    append: bool = False,
) -> None:
    filename = filename.format_map(variables)
    open_mode = 'w' + ('+' if append else '')
    with open(filename, open_mode) as fw:
        for line in body:
            fw.write(line.format_map(variables) + '\n')
    os.chmod(filename, int(mode, 8))


async def write_tempfile(
    variables: Mapping[str, Any],
    body: Iterable[str],
    mode: str = '644',
) -> Optional[str]:
    with tempfile.NamedTemporaryFile(
            'w', encoding='utf-8', suffix='.py', delete=False) as config:
        for line in body:
            config.write(line.format_map(variables))
    os.chmod(config.name, int(mode, 8))
    return config.name


async def run_command(
    variables: Mapping[str, Any],
    command: Iterable[str],
) -> Optional[MutableMapping[str, str]]:
    proc = await create_subprocess_exec(*(
        str(piece).format_map(variables) for piece in command
    ), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = await proc.communicate()
    return {'out': out.decode('utf8'), 'err': err.decode('utf8')}


async def mkdir(
    variables: Mapping[str, Any],
    path: str,
) -> None:
    Path(path.format_map(variables)).mkdir(parents=True, exist_ok=True)


async def log(
    variables: Mapping[str, Any],
    message: str,
    debug: bool = False,
) -> None:
    message = message.format_map(variables).replace('{', '{{').replace('}', '}}')
    if debug:
        logger.debug(message)
    else:
        logger.info(message)



import asyncio
import logging
import os
from pathlib import Path
import tempfile
from typing import List

from ... import BaseRunner

log = logging.getLogger()

DEFAULT_PYFLAGS: List[str] = []


class Runner(BaseRunner):

    log_prefix = 'h2o-kernel'
    default_runtime_path = '/opt/h2oai/dai/python/bin/python'
    default_child_env = {
        'TERM': 'xterm',
        'LANG': 'C.UTF-8',
        'SHELL': '/bin/bash',
        'USER': 'work',
        'HOME': '/home/work',
        'PATH': ':'.join([
            '/usr/local/nvidia/bin',
            '/usr/local/cuda/bin',
            '/usr/local/sbin',
            '/usr/local/bin',
            '/usr/sbin',
            '/usr/bin',
            '/sbin',
            '/bin',
        ]),
        'LD_LIBRARY_PATH': os.environ.get('LD_LIBRARY_PATH', ''),
        'LD_PRELOAD': os.environ.get('LD_PRELOAD', ''),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def init_with_loop(self):
        self.user_input_queue = asyncio.Queue()

        # Load H2O Daemon.
        print('Daemonizing H2O (run-dai.sh)...')
        Path('/opt/h2oai/dai').mkdir(parents=True, exist_ok=True)
        cmd = ['/opt/h2oai/dai/run-dai.sh']
        await self.run_subproc(cmd)

    async def build_heuristic(self) -> int:
        if Path('setup.py').is_file():
            cmd = [
                str(self.runtime_path), *DEFAULT_PYFLAGS,
                '-m', 'pip', 'install', '--user', '-e', '.',
            ]
            return await self.run_subproc(cmd)
        else:
            log.warning('skipping the build phase due to missing "setup.py" file')
            return 0

    async def execute_heuristic(self) -> int:
        if Path('main.py').is_file():
            cmd = [
                str(self.runtime_path), *DEFAULT_PYFLAGS,
                'main.py',
            ]
            return await self.run_subproc(cmd)
        else:
            log.error('cannot find the main script ("main.py").')
            return 127

    async def start_service(self, service_info):
        if service_info['name'] in ['jupyter', 'jupyterlab']:
            with tempfile.NamedTemporaryFile(
                    'w', encoding='utf-8', suffix='.py', delete=False) as config:
                print('c.NotebookApp.allow_root = True', file=config)
                print('c.NotebookApp.ip = "0.0.0.0"', file=config)
                print('c.NotebookApp.port = {}'.format(service_info['port']), file=config)
                print('c.NotebookApp.token = ""', file=config)
                print('c.FileContentsManager.delete_to_trash = False', file=config)
                print('c.NotebookApp.tornado_settings = {\'ws_ping_interval\': 10000}', file=config)
            jupyter_service_type = service_info['name']
            if jupyter_service_type == 'jupyter':
                jupyter_service_type = 'notebook'
            return [
                self.runtime_path, '-m', jupyter_service_type,
                '--no-browser',
                '--config', config.name,
            ], {}
        elif 'h2o' in service_info['name']:
            return ['echo', 'h2o daemon already started'], {}
        elif service_info['name'] == 'sftp':
            return [
                self.runtime_path,
                '-m', 'sftpserver',
                '--port', str(service_info['port']),
            ], {}



import json
from pathlib import Path
import sys


def extract_dotfiles():
    try:
        with open('/home/config/dotfiles.json') as fr:
            dotfiles = json.loads(fr.read())
    except FileNotFoundError:
        return
    work_dir = Path('/home/work')
    for dotfile in dotfiles:
        file_path = work_dir / dotfile['path']
        try:
            file_path.parent.mkdir(parents=True, exist_ok=True)
            file_path.write_text(dotfile['data'])
        except IOError:
            print(f"failed to write dotfile: {file_path}", file=sys.stderr)
        try:
            tmp = Path(file_path)
            while tmp != work_dir:
                tmp.chmod(int(dotfile['perm'], 8))
                tmp = tmp.parent
        except IOError:
            print(f"failed to chmod dotfile: {file_path}", file=sys.stderr)


if __name__ == '__main__':
    extract_dotfiles()



class InitializationError(Exception):
    """
    Errors during agent initialization and compute plugin setup
    """
    pass


class ResourceError(ValueError):
    pass


class UnsupportedResource(ResourceError):
    pass


class InvalidResourceCombination(ResourceError):
    pass


class InvalidResourceArgument(ResourceError):
    pass


class NotMultipleOfQuantum(InvalidResourceArgument):
    pass


class InsufficientResource(ResourceError):
    pass


class UnsupportedBaseDistroError(RuntimeError):
    pass


class K8sError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class AgentError(RuntimeError):
    '''
    A dummy exception class to distinguish agent-side errors passed via
    aiozmq.rpc calls.

    It carrise two args tuple: the exception type and exception arguments from
    the agent.
    '''

    def __init__(self, *args, exc_repr: str = None):
        super().__init__(*args)
        self.exc_repr = exc_repr



from __future__ import annotations

import asyncio
import base64
from decimal import Decimal
from functools import partial
from io import StringIO
import json
import logging
import os
from pathlib import Path
from aiohttp import web
import pkg_resources
import secrets
import shutil
import signal
import struct
from subprocess import CalledProcessError
import sys
from typing import (
    Any,
    FrozenSet,
    Dict,
    List,
    Literal,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    TYPE_CHECKING,
    Union,
)

from aiodocker.docker import Docker, DockerContainer
from aiodocker.exceptions import DockerError
import aiotools
from async_timeout import timeout
import zmq

from ai.backend.common.docker import (
    ImageRef,
    MIN_KERNELSPEC,
    MAX_KERNELSPEC,
)
from ai.backend.common.exception import ImageNotAvailable
from ai.backend.common.logging import BraceStyleAdapter, pretty
from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
from ai.backend.common.types import (
    AutoPullBehavior,
    BinarySize,
    ClusterInfo,
    ImageRegistry,
    KernelCreationConfig,
    KernelId,
    ContainerId,
    DeviceName,
    SlotName,
    MountPermission,
    MountTypes,
    ResourceSlot,
    Sentinel,
    current_resource_slots,
)
from ai.backend.common.utils import AsyncFileWriter, current_loop
from .kernel import DockerKernel, prepare_kernel_metadata_uri_handling
from .metadata.server import create_server as create_metadata_server
from .resources import detect_resources
from .utils import PersistentServiceContainer
from ..exception import UnsupportedResource, InitializationError
from ..fs import create_scratch_filesystem, destroy_scratch_filesystem
from ..kernel import KernelFeatures
from ..resources import (
    Mount,
    KernelResourceSpec,
)
from ..agent import (
    AbstractAgent,
    AbstractKernelCreationContext,
    ACTIVE_STATUS_SET,
    ComputerContext,
)
from ..proxy import proxy_connection, DomainSocketProxy
from ..resources import (
    AbstractComputePlugin,
    known_slot_types,
)
from ..server import (
    get_extra_volumes,
)
from ..types import (
    Container,
    Port,
    ContainerStatus,
    LifecycleEvent,
)
from ..utils import (
    closing_async,
    update_nested_dict,
    get_kernel_id_from_container,
    host_pid_to_container_pid,
    container_pid_to_host_pid,
)

if TYPE_CHECKING:
    from ai.backend.common.etcd import AsyncEtcd

log = BraceStyleAdapter(logging.getLogger(__name__))
eof_sentinel = Sentinel.TOKEN


def container_from_docker_container(src: DockerContainer) -> Container:
    ports = []
    for private_port, host_ports in src['NetworkSettings']['Ports'].items():
        private_port = int(private_port.split('/')[0])
        if host_ports is None:
            host_ip = '127.0.0.1'
            host_port = 0
        else:
            host_ip = host_ports[0]['HostIp']
            host_port = int(host_ports[0]['HostPort'])
        ports.append(Port(host_ip, private_port, host_port))
    return Container(
        id=src._id,
        status=src['State']['Status'],
        image=src['Config']['Image'],
        labels=src['Config']['Labels'],
        ports=ports,
        backend_obj=src,
    )


def _DockerError_reduce(self):
    return (
        type(self),
        (self.status, {'message': self.message}, *self.args),
    )


def _DockerContainerError_reduce(self):
    return (
        type(self),
        (self.status, {'message': self.message}, self.container_id, *self.args),
    )


class DockerKernelCreationContext(AbstractKernelCreationContext[DockerKernel]):

    scratch_dir: Path
    tmp_dir: Path
    config_dir: Path
    work_dir: Path
    container_configs: List[Mapping[str, Any]]
    domain_socket_proxies: List[DomainSocketProxy]
    computer_docker_args: Dict[str, Any]
    port_pool: Set[int]
    agent_sockpath: Path
    resource_lock: asyncio.Lock

    def __init__(
        self,
        kernel_id: KernelId,
        kernel_config: KernelCreationConfig,
        local_config: Mapping[str, Any],
        computers: MutableMapping[str, ComputerContext],
        port_pool: Set[int],
        agent_sockpath: Path,
        resource_lock: asyncio.Lock,
        restarting: bool = False,
    ) -> None:
        super().__init__(kernel_id, kernel_config, local_config, computers, restarting=restarting)
        scratch_dir = (self.local_config['container']['scratch-root'] / str(kernel_id)).resolve()
        tmp_dir = (self.local_config['container']['scratch-root'] / f'{kernel_id}_tmp').resolve()

        self.scratch_dir = scratch_dir
        self.tmp_dir = tmp_dir
        self.config_dir = scratch_dir / 'config'
        self.work_dir = scratch_dir / 'work'

        self.port_pool = port_pool
        self.agent_sockpath = agent_sockpath
        self.resource_lock = resource_lock

        self.container_configs = []
        self.domain_socket_proxies = []
        self.computer_docker_args  = {}

    def _kernel_resource_spec_read(self, filename):
        with open(filename, 'r') as f:
            resource_spec = KernelResourceSpec.read_from_file(f)
        return resource_spec

    async def get_extra_envs(self) -> Mapping[str, str]:
        return {}

    async def prepare_resource_spec(self) -> Tuple[KernelResourceSpec, Optional[Mapping[str, Any]]]:
        loop = current_loop()
        if self.restarting:
            resource_spec = await loop.run_in_executor(
                None,
                self._kernel_resource_spec_read,
                self.config_dir / 'resource.txt')
            resource_opts = None
        else:
            slots = ResourceSlot.from_json(self.kernel_config['resource_slots'])
            # Ensure that we have intrinsic slots.
            assert SlotName('cpu') in slots
            assert SlotName('mem') in slots
            # accept unknown slot type with zero values
            # but reject if they have non-zero values.
            for st, sv in slots.items():
                if st not in known_slot_types and sv != Decimal(0):
                    raise UnsupportedResource(st)
            # sanitize the slots
            current_resource_slots.set(known_slot_types)
            slots = slots.normalize_slots(ignore_unknown=True)
            resource_spec = KernelResourceSpec(
                container_id='',
                allocations={},
                slots={**slots},  # copy
                mounts=[],
                scratch_disk_size=0,  # TODO: implement (#70)
            )
            resource_opts = self.kernel_config.get('resource_opts', {})
        return resource_spec, resource_opts

    async def prepare_scratch(self) -> None:
        loop = current_loop()

        # Create the scratch, config, and work directories.
        if (
            sys.platform.startswith('linux')
            and self.local_config['container']['scratch-type'] == 'memory'
        ):
            await loop.run_in_executor(None, partial(self.tmp_dir.mkdir, exist_ok=True))
            await create_scratch_filesystem(self.scratch_dir, 64)
            await create_scratch_filesystem(self.tmp_dir, 64)
        else:
            await loop.run_in_executor(None, partial(self.scratch_dir.mkdir, exist_ok=True))

        def _create_scratch_dirs():
            self.config_dir.mkdir(parents=True, exist_ok=True)
            self.config_dir.chmod(0o755)
            self.work_dir.mkdir(parents=True, exist_ok=True)
            self.work_dir.chmod(0o755)

        await loop.run_in_executor(None, _create_scratch_dirs)

        if not self.restarting:
            # Since these files are bind-mounted inside a bind-mounted directory,
            # we need to touch them first to avoid their "ghost" files are created
            # as root in the host-side filesystem, which prevents deletion of scratch
            # directories when the agent is running as non-root.
            def _clone_dotfiles():
                jupyter_custom_css_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', 'jupyter-custom.css'))
                logo_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', 'logo.svg'))
                font_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', 'roboto.ttf'))
                font_italic_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', 'roboto-italic.ttf'))
                bashrc_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', '.bashrc'))
                bash_profile_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', '.bash_profile'))
                vimrc_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', '.vimrc'))
                tmux_conf_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', '.tmux.conf'))
                jupyter_custom_dir = (self.work_dir / '.jupyter' / 'custom')
                jupyter_custom_dir.mkdir(parents=True, exist_ok=True)
                shutil.copy(jupyter_custom_css_path.resolve(), jupyter_custom_dir / 'custom.css')
                shutil.copy(logo_path.resolve(), jupyter_custom_dir / 'logo.svg')
                shutil.copy(font_path.resolve(), jupyter_custom_dir / 'roboto.ttf')
                shutil.copy(font_italic_path.resolve(), jupyter_custom_dir / 'roboto-italic.ttf')
                shutil.copy(bashrc_path.resolve(), self.work_dir / '.bashrc')
                shutil.copy(bash_profile_path.resolve(), self.work_dir / '.bash_profile')
                shutil.copy(vimrc_path.resolve(), self.work_dir / '.vimrc')
                shutil.copy(tmux_conf_path.resolve(), self.work_dir / '.tmux.conf')
                if KernelFeatures.UID_MATCH in self.kernel_features:
                    uid = self.local_config['container']['kernel-uid']
                    gid = self.local_config['container']['kernel-gid']
                    if os.geteuid() == 0:  # only possible when I am root.
                        os.chown(self.work_dir, uid, gid)
                        os.chown(self.work_dir / '.jupyter', uid, gid)
                        os.chown(self.work_dir / '.jupyter' / 'custom', uid, gid)
                        os.chown(self.work_dir / '.bashrc', uid, gid)
                        os.chown(self.work_dir / '.bash_profile', uid, gid)
                        os.chown(self.work_dir / '.vimrc', uid, gid)
                        os.chown(self.work_dir / '.tmux.conf', uid, gid)

            await loop.run_in_executor(None, _clone_dotfiles)

    async def get_intrinsic_mounts(self) -> Sequence[Mount]:
        loop = current_loop()

        # scratch/config/tmp mounts
        mounts: List[Mount] = [
            Mount(MountTypes.BIND, self.config_dir, Path("/home/config"),
                  MountPermission.READ_ONLY),
            Mount(MountTypes.BIND, self.work_dir, Path("/home/work"),
                  MountPermission.READ_WRITE),
        ]
        if (sys.platform.startswith("linux") and
            self.local_config["container"]["scratch-type"] == "memory"):
            mounts.append(Mount(
                MountTypes.BIND,
                self.tmp_dir,
                Path("/tmp"),
                MountPermission.READ_WRITE,
            ))

        # lxcfs mounts
        lxcfs_root = Path("/var/lib/lxcfs")
        if lxcfs_root.is_dir():
            mounts.extend(
                Mount(
                    MountTypes.BIND,
                    lxcfs_proc_path,
                    "/" / lxcfs_proc_path.relative_to(lxcfs_root),
                    MountPermission.READ_WRITE,
                )
                for lxcfs_proc_path in (lxcfs_root / "proc").iterdir()
            )
            mounts.extend(
                Mount(
                    MountTypes.BIND,
                    lxcfs_root / path,
                    "/" / Path(path),
                    MountPermission.READ_WRITE,
                )
                for path in [
                    "sys/devices/system/cpu",
                    "sys/devices/system/cpu/online",
                ]
                if Path(lxcfs_root / path).exists()
            )

        # extra mounts
        async with closing_async(Docker()) as docker:
            extra_mount_list = await get_extra_volumes(docker, self.image_ref.short)
        mounts.extend(Mount(MountTypes.VOLUME, v.name, v.container_path, v.mode)
                      for v in extra_mount_list)

        # debug mounts
        if self.local_config['debug']['coredump']['enabled']:
            mounts.append(Mount(
                MountTypes.BIND,
                self.local_config['debug']['coredump']['path'],
                self.local_config['debug']['coredump']['core_path'],
                MountPermission.READ_WRITE,
            ))

        # agent-socket mount
        mounts.append(Mount(
            MountTypes.BIND,
            self.agent_sockpath,
            Path('/opt/kernel/agent.sock'),
            MountPermission.READ_WRITE,
        ))
        ipc_base_path = self.local_config['agent']['ipc-base-path']

        # domain-socket proxy mount
        # (used for special service containers such image importer)
        for host_sock_path in self.internal_data.get('domain_socket_proxies', []):
            await loop.run_in_executor(
                None,
                partial((ipc_base_path / 'proxy').mkdir, parents=True, exist_ok=True))
            host_proxy_path = ipc_base_path / 'proxy' / f'{secrets.token_hex(12)}.sock'
            proxy_server = await asyncio.start_unix_server(
                aiotools.apartial(proxy_connection, host_sock_path),
                str(host_proxy_path))
            await loop.run_in_executor(None, host_proxy_path.chmod, 0o666)
            self.domain_socket_proxies.append(DomainSocketProxy(
                Path(host_sock_path),
                host_proxy_path,
                proxy_server,
            ))
            mounts.append(Mount(
                MountTypes.BIND,
                host_proxy_path,
                host_sock_path,
                MountPermission.READ_WRITE,
            ))

        return mounts

    def resolve_krunner_filepath(self, filename) -> Path:
        return Path(pkg_resources.resource_filename(
            'ai.backend.agent', '../' + filename)).resolve()

    def get_runner_mount(
        self,
        type: MountTypes,
        src: Union[str, Path],
        target: Union[str, Path],
        perm: Literal['ro', 'rw'] = 'ro',
        opts: Mapping[str, Any] = None,
    ) -> Mount:
        return Mount(
            type, Path(src), Path(target),
            MountPermission(perm),
            opts=opts,
        )

    async def apply_network(self, cluster_info: ClusterInfo) -> None:
        if cluster_info['network_name'] is not None:
            self.container_configs.append({
                'HostConfig': {
                    'NetworkMode': cluster_info['network_name'],
                },
                'NetworkingConfig': {
                    'EndpointsConfig': {
                        cluster_info['network_name']: {
                            'Aliases': [self.kernel_config['cluster_hostname']],
                        },
                    },
                },
            })
            # RDMA mounts
            ib_root = Path("/dev/infiniband")
            if ib_root.is_dir() and (ib_root / "uverbs0").exists():
                self.container_configs.append({
                    'HostConfig': {
                        'Devices': [
                            {
                                'PathOnHost': "/dev/infiniband",
                                'PathInContainer': "/dev/infiniband",
                                'CgroupPermissions': "rwm",
                            },
                        ],
                    },
                })

    async def install_ssh_keypair(self, cluster_info: ClusterInfo) -> None:
        sshkey = cluster_info['ssh_keypair']
        if sshkey is None:
            return

        def _write_keypair():
            try:
                priv_key_path = (self.config_dir / 'ssh' / 'id_cluster')
                pub_key_path = (self.config_dir / 'ssh' / 'id_cluster.pub')
                priv_key_path.parent.mkdir(parents=True, exist_ok=True)
                priv_key_path.write_text(sshkey['private_key'])
                pub_key_path.write_text(sshkey['public_key'])
                if KernelFeatures.UID_MATCH in self.kernel_features:
                    uid = self.local_config['container']['kernel-uid']
                    gid = self.local_config['container']['kernel-gid']
                    if os.geteuid() == 0:  # only possible when I am root.
                        os.chown(str(priv_key_path), uid, gid)
                        os.chown(str(pub_key_path), uid, gid)
                priv_key_path.chmod(0o600)
            except Exception:
                log.exception('error while writing cluster keypair')

        current_loop().run_in_executor(None, _write_keypair)  # ???

    async def process_mounts(self, mounts: Sequence[Mount]):
        def fix_unsupported_perm(folder_perm: MountPermission) -> MountPermission:
            if folder_perm == MountPermission.RW_DELETE:
                # TODO: enforce readable/writable but not deletable
                # (Currently docker's READ_WRITE includes DELETE)
                return MountPermission.READ_WRITE
            return folder_perm

        container_config = {
            'HostConfig': {
                'Mounts': [
                    {
                        'Target': str(mount.target),
                        'Source': str(mount.source),
                        'Type': mount.type.value,
                        'ReadOnly': fix_unsupported_perm(mount.permission) == MountPermission.READ_ONLY,
                        f'{mount.type.value.capitalize()}Options':
                            mount.opts if mount.opts else {},
                    }
                    for mount in mounts
                ],
            },
        }
        self.container_configs.append(container_config)

    async def apply_accelerator_allocation(self, computer, device_alloc) -> None:
        async with closing_async(Docker()) as docker:
            update_nested_dict(
                self.computer_docker_args,
                await computer.generate_docker_args(docker, device_alloc),
            )

    async def spawn(
        self,
        resource_spec: KernelResourceSpec,
        resource_opts,
        environ: Mapping[str, str],
        service_ports,
        preopen_ports,
        cmdargs: List[str],
    ) -> DockerKernel:
        loop = current_loop()
        image_labels = self.kernel_config['image']['labels']
        exposed_ports = [2000, 2001]
        for sport in service_ports:
            exposed_ports.extend(sport['container_ports'])

        if self.restarting:
            pass
        else:
            # Create bootstrap.sh into workdir if needed
            if bootstrap := self.kernel_config.get('bootstrap_script'):

                def _write_user_bootstrap_script():
                    (self.work_dir / 'bootstrap.sh').write_text(bootstrap)
                    if KernelFeatures.UID_MATCH in self.kernel_features:
                        uid = self.local_config['container']['kernel-uid']
                        gid = self.local_config['container']['kernel-gid']
                        if os.geteuid() == 0:
                            os.chown(self.work_dir / 'bootstrap.sh', uid, gid)

                await loop.run_in_executor(None, _write_user_bootstrap_script)

            with StringIO() as buf:
                for k, v in environ.items():
                    buf.write(f'{k}={v}\n')
                accel_envs = self.computer_docker_args.get('Env', [])
                for env in accel_envs:
                    buf.write(f'{env}\n')
                await loop.run_in_executor(
                    None,
                    (self.config_dir / 'environ.txt').write_bytes,
                    buf.getvalue().encode('utf8'),
                )

            with StringIO() as buf:
                resource_spec.write_to_file(buf)
                for dev_type, device_alloc in resource_spec.allocations.items():
                    computer_self = self.computers[dev_type]
                    kvpairs = \
                        await computer_self.instance.generate_resource_data(device_alloc)
                    for k, v in kvpairs.items():
                        buf.write(f'{k}={v}\n')
                await loop.run_in_executor(
                    None,
                    (self.config_dir / 'resource.txt').write_bytes,
                    buf.getvalue().encode('utf8'),
                )

            docker_creds = self.internal_data.get('docker_credentials')
            if docker_creds:
                await loop.run_in_executor(
                    None,
                    (self.config_dir / 'docker-creds.json').write_text,
                    json.dumps(docker_creds))

        # TODO: refactor out dotfiles/sshkey initialization to the base agent?

        shutil.copyfile(self.config_dir / 'environ.txt', self.config_dir / 'environ_base.txt')
        shutil.copyfile(self.config_dir / 'resource.txt', self.config_dir / 'resource_base.txt')
        # Create SSH keypair only if ssh_keypair internal_data exists and
        # /home/work/.ssh folder is not mounted.
        if self.internal_data.get('ssh_keypair'):
            for mount in resource_spec.mounts:
                container_path = str(mount).split(':')[1]
                if container_path == '/home/work/.ssh':
                    break
            else:
                pubkey = self.internal_data['ssh_keypair']['public_key'].encode('ascii')
                privkey = self.internal_data['ssh_keypair']['private_key'].encode('ascii')
                ssh_dir = self.work_dir / '.ssh'

                def _populate_ssh_config():
                    ssh_dir.mkdir(parents=True, exist_ok=True)
                    ssh_dir.chmod(0o700)
                    (ssh_dir / 'authorized_keys').write_bytes(pubkey)
                    (ssh_dir / 'authorized_keys').chmod(0o600)
                    (self.work_dir / 'id_container').write_bytes(privkey)
                    (self.work_dir / 'id_container').chmod(0o600)
                    if KernelFeatures.UID_MATCH in self.kernel_features:
                        uid = self.local_config['container']['kernel-uid']
                        gid = self.local_config['container']['kernel-gid']
                        if os.geteuid() == 0:  # only possible when I am root.
                            os.chown(ssh_dir, uid, gid)
                            os.chown(ssh_dir / 'authorized_keys', uid, gid)
                            os.chown(self.work_dir / 'id_container', uid, gid)

                await loop.run_in_executor(None, _populate_ssh_config)

        # higher priority dotfiles are stored last to support overwriting
        for dotfile in self.internal_data.get('dotfiles', []):
            if dotfile['path'].startswith('/'):
                if dotfile['path'].startswith('/home/'):
                    path_arr = dotfile['path'].split('/')
                    file_path: Path = self.scratch_dir / '/'.join(path_arr[2:])
                else:
                    file_path = Path(dotfile['path'])
            else:
                file_path = self.work_dir / dotfile['path']
            file_path.parent.mkdir(parents=True, exist_ok=True)
            await loop.run_in_executor(
                None,
                file_path.write_text,
                dotfile['data'])

            tmp = Path(file_path)
            while tmp != self.work_dir:
                tmp.chmod(int(dotfile['perm'], 8))
                # only possible when I am root.
                if KernelFeatures.UID_MATCH in self.kernel_features and os.geteuid() == 0:
                    uid = self.local_config['container']['kernel-uid']
                    gid = self.local_config['container']['kernel-gid']
                    os.chown(tmp, uid, gid)
                tmp = tmp.parent

        # PHASE 4: Run!
        container_bind_host = self.local_config['container']['bind-host']
        advertised_kernel_host = self.local_config['container'].get('advertised-host')
        if len(exposed_ports) > len(self.port_pool):
            raise RuntimeError('Container ports are not sufficiently available.')
        host_ports = []
        for eport in exposed_ports:
            hport = self.port_pool.pop()
            host_ports.append(hport)

        container_log_size = self.local_config['agent']['container-logs']['max-length']
        container_log_file_count = 5
        container_log_file_size = BinarySize(container_log_size // container_log_file_count)
        container_config: MutableMapping[str, Any] = {
            'Image': self.image_ref.canonical,
            'Tty': True,
            'OpenStdin': True,
            'Privileged': False,
            'StopSignal': 'SIGINT',
            'ExposedPorts': {
                f'{port}/tcp': {} for port in exposed_ports
            },
            'EntryPoint': ["/opt/kernel/entrypoint.sh"],
            'Cmd': cmdargs,
            'Env': [f'{k}={v}' for k, v in environ.items()],
            'WorkingDir': "/home/work",
            'Hostname': self.kernel_config['cluster_hostname'],
            'Labels': {
                'ai.backend.kernel-id': str(self.kernel_id),
                'ai.backend.internal.block-service-ports':
                    '1' if self.internal_data.get('block_service_ports', False) else '0',
            },
            'HostConfig': {
                'Init': True,
                'PortBindings': {
                    f'{eport}/tcp': [{'HostPort': str(hport),
                                      'HostIp': str(container_bind_host)}]
                    for eport, hport in zip(exposed_ports, host_ports)
                },
                'PublishAllPorts': False,  # we manage port mapping manually!
                'CapAdd': [
                    'IPC_LOCK',  # for hugepages and RDMA
                ],
                'Ulimits': [
                    {"Name": "nofile", "Soft": 1048576, "Hard": 1048576},
                    {"Name": "memlock", "Soft": -1, "Hard": -1},
                ],
                'LogConfig': {
                    'Type': 'local',  # for efficient docker-specific storage
                    'Config': {
                        # these fields must be str
                        # (ref: https://docs.docker.com/config/containers/logging/local/)
                        'max-size': f"{container_log_file_size:s}",
                        'max-file': str(container_log_file_count),
                        'compress': 'false',
                    },
                },
            },
        }
        # merge all container configs generated during prior preparation steps
        for c in self.container_configs:
            update_nested_dict(container_config, c)
        if self.local_config['container']['sandbox-type'] == 'jail':
            update_nested_dict(container_config, {
                'HostConfig': {
                    'SecurityOpt': ['seccomp=unconfined', 'apparmor=unconfined'],
                },
            })

        if resource_opts and resource_opts.get('shmem'):
            shmem = int(resource_opts.get('shmem', '0'))
            self.computer_docker_args['HostConfig']['ShmSize'] = shmem
            self.computer_docker_args['HostConfig']['MemorySwap'] -= shmem
            self.computer_docker_args['HostConfig']['Memory'] -= shmem

        encoded_preopen_ports = ','.join(f'{port_no}:preopen:{port_no}' for port_no in preopen_ports)
        container_config['Labels']['ai.backend.service-ports'] = \
                image_labels['ai.backend.service-ports'] + ',' + encoded_preopen_ports
        update_nested_dict(container_config, self.computer_docker_args)
        kernel_name = f"kernel.{self.image_ref.name.split('/')[-1]}.{self.kernel_id}"
        if self.local_config['debug']['log-kernel-config']:
            log.debug('full container config: {!r}', pretty(container_config))

        # optional local override of docker config
        extra_container_opts_name = 'agent-docker-container-opts.json'
        for extra_container_opts_file in [
            Path('/etc/backend.ai') / extra_container_opts_name,
            Path.home() / '.config' / 'backend.ai' / extra_container_opts_name,
            Path.cwd() / extra_container_opts_name,
        ]:
            if extra_container_opts_file.is_file():
                try:
                    extra_container_opts = json.loads(extra_container_opts_file.read_bytes())
                    update_nested_dict(container_config, extra_container_opts)
                except IOError:
                    pass

        # We are all set! Create and start the container.
        async with closing_async(Docker()) as docker:
            try:
                container = await docker.containers.create(
                    config=container_config, name=kernel_name)
                cid = container._id

                resource_spec.container_id = cid
                # Write resource.txt again to update the contaienr id.
                with open(self.config_dir / 'resource.txt', 'w') as f:
                    await loop.run_in_executor(None, resource_spec.write_to_file, f)
                async with AsyncFileWriter(
                    target_filename=self.config_dir / 'resource.txt',
                    access_mode='a',
                ) as writer:
                    for dev_name, device_alloc in resource_spec.allocations.items():
                        computer_ctx = self.computers[dev_name]
                        kvpairs = \
                            await computer_ctx.instance.generate_resource_data(device_alloc)
                        for k, v in kvpairs.items():
                            await writer.write(f'{k}={v}\n')

                await container.start()
            except asyncio.CancelledError:
                raise
            except Exception:
                # Oops, we have to restore the allocated resources!
                if (sys.platform.startswith('linux') and
                    self.local_config['container']['scratch-type'] == 'memory'):
                    await destroy_scratch_filesystem(self.scratch_dir)
                    await destroy_scratch_filesystem(self.tmp_dir)
                    await loop.run_in_executor(None, shutil.rmtree, self.tmp_dir)
                await loop.run_in_executor(None, shutil.rmtree, self.scratch_dir)
                self.port_pool.update(host_ports)
                async with self.resource_lock:
                    for dev_name, device_alloc in resource_spec.allocations.items():
                        self.computers[dev_name].alloc_map.free(device_alloc)
                raise

            ctnr_host_port_map: MutableMapping[int, int] = {}
            stdin_port = 0
            stdout_port = 0
            for idx, port in enumerate(exposed_ports):
                host_port = int((await container.port(port))[0]['HostPort'])
                assert host_port == host_ports[idx]
                if port == 2000:     # intrinsic
                    repl_in_port = host_port
                elif port == 2001:   # intrinsic
                    repl_out_port = host_port
                elif port == 2002:   # legacy
                    stdin_port = host_port
                elif port == 2003:   # legacy
                    stdout_port = host_port
                else:
                    ctnr_host_port_map[port] = host_port
            for sport in service_ports:
                sport['host_ports'] = tuple(
                    ctnr_host_port_map[cport] for cport in sport['container_ports']
                )

        kernel_obj = await DockerKernel.new(
            self.kernel_id,
            self.image_ref,
            self.kspec_version,
            agent_config=self.local_config,
            service_ports=service_ports,
            resource_spec=resource_spec,
            environ=environ,
            data={
                'container_id': container._id,
                'kernel_host': advertised_kernel_host or container_bind_host,
                'repl_in_port': repl_in_port,
                'repl_out_port': repl_out_port,
                'stdin_port': stdin_port,    # legacy
                'stdout_port': stdout_port,  # legacy
                'host_ports': host_ports,
                'domain_socket_proxies': self.domain_socket_proxies,
                'block_service_ports': self.internal_data.get('block_service_ports', False),
            })
        return kernel_obj


class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):

    monitor_docker_task: asyncio.Task
    agent_sockpath: Path
    agent_sock_task: asyncio.Task
    scan_images_timer: asyncio.Task
    metadata_server_runner: web.AppRunner
    docker_ptask_group: aiotools.PersistentTaskGroup

    def __init__(
        self,
        etcd: AsyncEtcd,
        local_config: Mapping[str, Any],
        *,
        stats_monitor: StatsPluginContext,
        error_monitor: ErrorPluginContext,
        skip_initial_scan: bool = False,
    ) -> None:
        super().__init__(
            etcd,
            local_config,
            stats_monitor=stats_monitor,
            error_monitor=error_monitor,
            skip_initial_scan=skip_initial_scan,
        )

    async def __ainit__(self) -> None:
        async with closing_async(Docker()) as docker:
            if not self._skip_initial_scan:
                docker_version = await docker.version()
                log.info('running with Docker {0} with API {1}',
                         docker_version['Version'], docker_version['ApiVersion'])
        await super().__ainit__()
        await self.check_swarm_status()
        if self.heartbeat_extra_info['swarm_enabled']:
            log.info('The Docker Swarm cluster is configured and enabled')
        ipc_base_path = self.local_config['agent']['ipc-base-path']
        (ipc_base_path / 'container').mkdir(parents=True, exist_ok=True)
        self.agent_sockpath = ipc_base_path / 'container' / f'agent.{self.local_instance_id}.sock'
        socket_relay_name = f"backendai-socket-relay.{self.local_instance_id}"
        socket_relay_container = PersistentServiceContainer(
            'backendai-socket-relay:latest',
            {
                'Cmd': [
                    f"UNIX-LISTEN:/ipc/{self.agent_sockpath.name},unlink-early,fork,mode=777",
                    f"TCP-CONNECT:127.0.0.1:{self.local_config['agent']['agent-sock-port']}",
                ],
                'HostConfig': {
                    'Mounts': [
                        {
                            'Type': 'bind',
                            'Source': str(ipc_base_path / 'container'),
                            'Target': '/ipc',
                        },
                    ],
                    'NetworkMode': 'host',
                },
            },
            name=socket_relay_name,
        )
        await socket_relay_container.ensure_running_latest()
        self.agent_sock_task = asyncio.create_task(self.handle_agent_socket())
        self.monitor_docker_task = asyncio.create_task(self.monitor_docker_events())
        self.monitor_swarm_task = asyncio.create_task(self.check_swarm_status(as_task=True))
        self.docker_ptask_group = aiotools.PersistentTaskGroup()

        await prepare_kernel_metadata_uri_handling(self.local_config)
        metadata_server_runner = web.AppRunner(
            await create_metadata_server(self.local_config, self.kernel_registry),
        )
        await metadata_server_runner.setup()
        site = web.TCPSite(metadata_server_runner, '0.0.0.0', 40128)
        await site.start()
        self.metadata_server_runner = metadata_server_runner
        # For legacy accelerator plugins
        self.docker = Docker()

    async def shutdown(self, stop_signal: signal.Signals):
        # Stop handling agent sock.
        if self.agent_sock_task is not None:
            self.agent_sock_task.cancel()
            await self.agent_sock_task
        if self.docker_ptask_group is not None:
            await self.docker_ptask_group.shutdown()

        try:
            await super().shutdown(stop_signal)
        finally:
            # Stop docker event monitoring.
            if self.monitor_docker_task is not None:
                self.monitor_docker_task.cancel()
                await self.monitor_docker_task

        if self.monitor_swarm_task is not None:
            self.monitor_swarm_task.cancel()
            await self.monitor_swarm_task

        await self.metadata_server_runner.cleanup()
        if self.docker:
            await self.docker.close()

    async def detect_resources(self) -> Tuple[
        Mapping[DeviceName, AbstractComputePlugin],
        Mapping[SlotName, Decimal],
    ]:
        return await detect_resources(self.etcd, self.local_config)

    async def enumerate_containers(
        self,
        status_filter: FrozenSet[ContainerStatus] = ACTIVE_STATUS_SET,
    ) -> Sequence[Tuple[KernelId, Container]]:
        result = []
        fetch_tasks = []
        async with closing_async(Docker()) as docker:
            for container in (await docker.containers.list()):

                async def _fetch_container_info(container):
                    kernel_id = "(unknown)"
                    try:
                        kernel_id = await get_kernel_id_from_container(container)
                        if kernel_id is None:
                            return
                        if container['State']['Status'] in status_filter:
                            await container.show()
                            result.append(
                                (
                                    kernel_id,
                                    container_from_docker_container(container),
                                ),
                            )
                    except asyncio.CancelledError:
                        pass
                    except Exception:
                        log.exception(
                            "error while fetching container information (cid:{}, k:{})",
                            container._id, kernel_id,
                        )

                fetch_tasks.append(_fetch_container_info(container))

            await asyncio.gather(*fetch_tasks, return_exceptions=True)
        return result

    async def check_swarm_status(self, as_task=False):
        try:
            while True:
                if as_task:
                    await asyncio.sleep(30)
                try:
                    swarm_enabled = self.local_config['container'].get('swarm-enabled', False)
                    if not swarm_enabled:
                        continue
                    async with closing_async(Docker()) as docker:
                        docker_info = await docker.system.info()
                        if docker_info['Swarm']['LocalNodeState'] == 'inactive':
                            raise InitializationError(
                                "The swarm mode is enabled but the node state of "
                                "the local Docker daemon is inactive.",
                            )
                except InitializationError as e:
                    log.exception(str(e))
                    swarm_enabled = False
                finally:
                    self.heartbeat_extra_info = {
                        'swarm_enabled': swarm_enabled,
                    }
                    if not as_task:
                        return
        except asyncio.CancelledError:
            pass

    async def scan_images(self) -> Mapping[str, str]:
        async with closing_async(Docker()) as docker:
            all_images = await docker.images.list()
            updated_images = {}
            for image in all_images:
                if image['RepoTags'] is None:
                    continue
                for repo_tag in image['RepoTags']:
                    if repo_tag.endswith('<none>'):
                        continue
                    img_detail = await docker.images.inspect(repo_tag)
                    labels = img_detail['Config']['Labels']
                    if labels is None or 'ai.backend.kernelspec' not in labels:
                        continue
                    kernelspec = int(labels['ai.backend.kernelspec'])
                    if MIN_KERNELSPEC <= kernelspec <= MAX_KERNELSPEC:
                        updated_images[repo_tag] = img_detail['Id']
            for added_image in (updated_images.keys() - self.images.keys()):
                log.debug('found kernel image: {0}', added_image)
            for removed_image in (self.images.keys() - updated_images.keys()):
                log.debug('removed kernel image: {0}', removed_image)
            return updated_images

    async def handle_agent_socket(self):
        """
        A simple request-reply socket handler for in-container processes.
        For ease of implementation in low-level languages such as C,
        it uses a simple C-friendly ZeroMQ-based multipart messaging protocol.

        The agent listens on a local TCP port and there is a socat relay
        that proxies this port via a UNIX domain socket mounted inside
        actual containers.  The reason for this is to avoid inode changes
        upon agent restarts by keeping the relay container running persistently,
        so that the mounted UNIX socket files don't get to refere a dangling pointer
        when the agent is restarted.

        Request message:
            The first part is the requested action as string,
            The second part and later are arguments.

        Reply message:
            The first part is a 32-bit integer (int in C)
                (0: success)
                (-1: generic unhandled error)
                (-2: invalid action)
            The second part and later are arguments.

        All strings are UTF-8 encoded.
        """
        terminating = False
        while True:
            agent_sock = self.zmq_ctx.socket(zmq.REP)
            try:
                agent_sock.bind(f"tcp://127.0.0.1:{self.local_config['agent']['agent-sock-port']}")
                while True:
                    msg = await agent_sock.recv_multipart()
                    if not msg:
                        break
                    try:
                        if msg[0] == b'host-pid-to-container-pid':
                            container_id = msg[1].decode()
                            host_pid = struct.unpack('i', msg[2])[0]
                            container_pid = await host_pid_to_container_pid(
                                container_id, host_pid)
                            reply = [
                                struct.pack('i', 0),
                                struct.pack('i', container_pid),
                            ]
                        elif msg[0] == b'container-pid-to-host-pid':
                            container_id = msg[1].decode()
                            container_pid = struct.unpack('i', msg[2])[0]
                            host_pid = await container_pid_to_host_pid(
                                container_id, container_pid)
                            reply = [
                                struct.pack('i', 0),
                                struct.pack('i', host_pid),
                            ]
                        else:
                            reply = [struct.pack('i', -2), b'Invalid action']
                    except asyncio.CancelledError:
                        terminating = True
                        raise
                    except Exception as e:
                        log.exception("handle_agent_socket(): internal error")
                        reply = [struct.pack('i', -1), f'Error: {e}'.encode('utf-8')]
                    await agent_sock.send_multipart(reply)
            except asyncio.CancelledError:
                terminating = True
                return
            except zmq.ZMQError:
                log.exception("handle_agent_socket(): zmq error")
            finally:
                agent_sock.close()
                if not terminating:
                    log.info("handle_agent_socket(): rebinding the socket")

    async def pull_image(self, image_ref: ImageRef, registry_conf: ImageRegistry) -> None:
        auth_config = None
        reg_user = registry_conf.get('username')
        reg_passwd = registry_conf.get('password')
        if reg_user and reg_passwd:
            encoded_creds = base64.b64encode(
                f'{reg_user}:{reg_passwd}'.encode('utf-8')) \
                .decode('ascii')
            auth_config = {
                'auth': encoded_creds,
            }
        log.info('pulling image {} from registry', image_ref.canonical)
        async with closing_async(Docker()) as docker:
            await docker.images.pull(
                image_ref.canonical,
                auth=auth_config)

    async def check_image(self, image_ref: ImageRef, image_id: str, auto_pull: AutoPullBehavior) -> bool:
        try:
            async with closing_async(Docker()) as docker:
                image_info = await docker.images.inspect(image_ref.canonical)
                if auto_pull == AutoPullBehavior.DIGEST:
                    if image_info['Id'] != image_id:
                        return True
            log.info('found the local up-to-date image for {}', image_ref.canonical)
        except DockerError as e:
            if e.status == 404:
                if auto_pull == AutoPullBehavior.DIGEST:
                    return True
                elif auto_pull == AutoPullBehavior.TAG:
                    return True
                elif auto_pull == AutoPullBehavior.NONE:
                    raise ImageNotAvailable(image_ref)
            else:
                raise
        return False

    async def init_kernel_context(
        self,
        kernel_id: KernelId,
        kernel_config: KernelCreationConfig,
        *,
        restarting: bool = False,
    ) -> DockerKernelCreationContext:
        return DockerKernelCreationContext(
            kernel_id,
            kernel_config,
            self.local_config,
            self.computers,
            self.port_pool,
            self.agent_sockpath,
            self.resource_lock,
            restarting=restarting,
        )

    async def restart_kernel__load_config(
        self,
        kernel_id: KernelId,
        name: str,
    ) -> bytes:
        loop = current_loop()
        scratch_dir = (self.local_config['container']['scratch-root'] / str(kernel_id)).resolve()
        config_dir = scratch_dir / 'config'
        return await loop.run_in_executor(
            None,
            (config_dir / name).read_bytes,
        )

    async def restart_kernel__store_config(
        self,
        kernel_id: KernelId,
        name: str,
        data: bytes,
    ) -> None:
        loop = current_loop()
        scratch_dir = (self.local_config['container']['scratch-root'] / str(kernel_id)).resolve()
        config_dir = scratch_dir / 'config'
        return await loop.run_in_executor(
            None,
            (config_dir / name).write_bytes,
            data,
        )

    async def destroy_kernel(
        self,
        kernel_id: KernelId,
        container_id: Optional[ContainerId],
    ) -> None:
        if container_id is None:
            return
        try:
            async with closing_async(Docker()) as docker:
                container = docker.containers.container(container_id)
                # The default timeout of the docker stop API is 10 seconds
                # to kill if container does not self-terminate.
                await container.stop()
        except DockerError as e:
            if e.status == 409 and 'is not running' in e.message:
                # already dead
                log.warning('destroy_kernel(k:{0}) already dead', kernel_id)
                await self.rescan_resource_usage()
            elif e.status == 404:
                # missing
                log.warning('destroy_kernel(k:{0}) kernel missing, '
                            'forgetting this kernel', kernel_id)
                await self.rescan_resource_usage()
            else:
                log.exception('destroy_kernel(k:{0}) kill error', kernel_id)
                self.error_monitor.capture_exception()

    async def clean_kernel(
        self,
        kernel_id: KernelId,
        container_id: Optional[ContainerId],
        restarting: bool,
    ) -> None:
        loop = current_loop()
        async with closing_async(Docker()) as docker:
            if container_id is not None:
                container = docker.containers.container(container_id)

                async def log_iter():
                    it = container.log(
                        stdout=True, stderr=True, follow=True,
                    )
                    async with aiotools.aclosing(it):
                        async for line in it:
                            yield line.encode('utf-8')

                try:
                    with timeout(60):
                        await self.collect_logs(kernel_id, container_id, log_iter())
                except asyncio.TimeoutError:
                    log.warning('timeout for collecting container logs (k:{}, cid:{})',
                                kernel_id, container_id)
                except Exception as e:
                    log.warning('error while collecting container logs (k:{}, cid:{})',
                                kernel_id, container_id, exc_info=e)

            kernel_obj = self.kernel_registry.get(kernel_id)
            if kernel_obj is not None:
                for domain_socket_proxy in kernel_obj.get('domain_socket_proxies', []):
                    if domain_socket_proxy.proxy_server.is_serving():
                        domain_socket_proxy.proxy_server.close()
                        await domain_socket_proxy.proxy_server.wait_closed()
                        try:
                            domain_socket_proxy.host_proxy_path.unlink()
                        except IOError:
                            pass

            if not self.local_config['debug']['skip-container-deletion'] and container_id is not None:
                container = docker.containers.container(container_id)
                try:
                    with timeout(90):
                        await container.delete(force=True, v=True)
                except DockerError as e:
                    if e.status == 409 and 'already in progress' in e.message:
                        return
                    elif e.status == 404:
                        return
                    else:
                        log.exception(
                            'unexpected docker error while deleting container (k:{}, c:{})',
                            kernel_id, container_id)
                except asyncio.TimeoutError:
                    log.warning('container deletion timeout (k:{}, c:{})',
                                kernel_id, container_id)

            if not restarting:
                scratch_root = self.local_config['container']['scratch-root']
                scratch_dir = scratch_root / str(kernel_id)
                tmp_dir = scratch_root / f'{kernel_id}_tmp'
                try:
                    if (sys.platform.startswith('linux') and
                        self.local_config['container']['scratch-type'] == 'memory'):
                        await destroy_scratch_filesystem(scratch_dir)
                        await destroy_scratch_filesystem(tmp_dir)
                        await loop.run_in_executor(None, shutil.rmtree, tmp_dir)
                    await loop.run_in_executor(None, shutil.rmtree, scratch_dir)
                except CalledProcessError:
                    pass
                except FileNotFoundError:
                    pass

    async def create_overlay_network(self, network_name: str) -> None:
        if not self.heartbeat_extra_info['swarm_enabled']:
            raise RuntimeError("This agent has not joined to a swarm cluster.")
        async with closing_async(Docker()) as docker:
            await docker.networks.create({
                'Name': network_name,
                'Driver': 'overlay',
                'Attachable': True,
                'Labels': {
                    'ai.backend.cluster-network': '1',
                },
            })

    async def destroy_overlay_network(self, network_name: str) -> None:
        docker = Docker()
        try:
            network = await docker.networks.get(network_name)
            await network.delete()
        finally:
            await docker.close()

    async def create_local_network(self, network_name: str) -> None:
        async with closing_async(Docker()) as docker:
            await docker.networks.create({
                'Name': network_name,
                'Driver': 'bridge',
                'Labels': {
                    'ai.backend.cluster-network': '1',
                },
            })

    async def destroy_local_network(self, network_name: str) -> None:
        async with closing_async(Docker()) as docker:
            network = await docker.networks.get(network_name)
            await network.delete()

    async def monitor_docker_events(self):

        async def handle_action_start(kernel_id: KernelId, evdata: Mapping[str, Any]) -> None:
            await self.inject_container_lifecycle_event(
                kernel_id,
                LifecycleEvent.START,
                'new-container-started',
                container_id=ContainerId(evdata['Actor']['ID']),
            )

        async def handle_action_die(kernel_id: KernelId, evdata: Mapping[str, Any]) -> None:
            # When containers die, we immediately clean up them.
            reason = None
            kernel_obj = self.kernel_registry.get(kernel_id)
            if kernel_obj is not None:
                reason = kernel_obj.termination_reason
            try:
                exit_code = evdata['Actor']['Attributes']['exitCode']
            except KeyError:
                exit_code = 255
            await self.inject_container_lifecycle_event(
                kernel_id,
                LifecycleEvent.CLEAN,
                reason or 'self-terminated',
                container_id=ContainerId(evdata['Actor']['ID']),
                exit_code=exit_code,
            )

        while True:
            async with closing_async(Docker()) as docker:
                subscriber = docker.events.subscribe(create_task=True)
                try:
                    while True:
                        try:
                            # ref: https://docs.docker.com/engine/api/v1.40/#operation/SystemEvents
                            evdata = await subscriber.get()
                            if evdata is None:
                                # Break out to the outermost loop when the connection is closed
                                log.info(
                                    "monitor_docker_events(): "
                                    "restarting aiodocker event subscriber",
                                )
                                break
                            if evdata['Type'] != 'container':
                                # Our interest is the container-related events
                                continue
                            container_name = evdata['Actor']['Attributes']['name']
                            kernel_id = await get_kernel_id_from_container(container_name)
                            if kernel_id is None:
                                continue
                            if (
                                self.local_config['debug']['log-docker-events']
                                and evdata['Action'] in ('start', 'die')
                            ):
                                log.debug('docker-event: action={}, actor={}',
                                          evdata['Action'], evdata['Actor'])
                            if evdata['Action'] == 'start':
                                await asyncio.shield(self.docker_ptask_group.create_task(
                                    handle_action_start(kernel_id, evdata),
                                ))
                            elif evdata['Action'] == 'die':
                                await asyncio.shield(self.docker_ptask_group.create_task(
                                    handle_action_die(kernel_id, evdata),
                                ))
                        except asyncio.CancelledError:
                            # We are shutting down...
                            return
                        except Exception:
                            log.exception("monitor_docker_events(): unexpected error")
                finally:
                    await asyncio.shield(self.docker_ptask_group.create_task(
                        docker.events.stop(),
                    ))



"""
A module to collect various performance metrics of Docker containers.

Reference: https://www.datadoghq.com/blog/how-to-collect-docker-metrics/
"""

import asyncio
from decimal import Decimal
import enum
import logging
import sys
import time
from typing import (
    Callable,
    Dict,
    FrozenSet,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    TYPE_CHECKING,
)
import aioredis

import attr

from ai.backend.common import redis
from ai.backend.common.identity import is_containerized
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common import msgpack
from ai.backend.common.types import (
    ContainerId, DeviceId, KernelId,
    MetricKey, MetricValue, MovingStatValue,
)
from .utils import (
    remove_exponent,
)
if TYPE_CHECKING:
    from .agent import AbstractAgent

__all__ = (
    'StatContext',
    'StatModes',
    'MetricTypes',
    'NodeMeasurement',
    'ContainerMeasurement',
    'Measurement',
)

log = BraceStyleAdapter(logging.getLogger('ai.backend.agent.stats'))


def check_cgroup_available():
    """
    Check if the host OS provides cgroups.
    """
    return (not is_containerized() and sys.platform.startswith('linux'))


class StatModes(enum.Enum):
    CGROUP = 'cgroup'
    DOCKER = 'docker'

    @staticmethod
    def get_preferred_mode():
        """
        Returns the most preferred statistics collector type for the host OS.
        """
        if check_cgroup_available():
            return StatModes.CGROUP
        return StatModes.DOCKER


class MetricTypes(enum.Enum):
    USAGE = 0        # for instant snapshot (e.g., used memory bytes, used cpu msec)
    RATE = 1         # for rate of increase (e.g., I/O bps)
    UTILIZATION = 2  # for ratio of resource occupation time per measurement interval (e.g., CPU util)
    ACCUMULATED = 3  # for accumulated value (e.g., total number of events)


@attr.s(auto_attribs=True, slots=True)
class Measurement:
    value: Decimal
    capacity: Optional[Decimal] = None


@attr.s(auto_attribs=True, slots=True)
class NodeMeasurement:
    """
    Collection of per-node and per-agent statistics for a specific metric.
    """
    # 2-tuple of Decimals mean raw values for (usage, available)
    # Percent values are calculated from them.
    key: str
    type: MetricTypes
    per_node: Measurement
    per_device: Mapping[DeviceId, Measurement] = attr.Factory(dict)
    unit_hint: Optional[str] = None
    stats_filter: FrozenSet[str] = attr.Factory(frozenset)
    current_hook: Optional[Callable[['Metric'], Decimal]] = None


@attr.s(auto_attribs=True, slots=True)
class ContainerMeasurement:
    """
    Collection of per-container statistics for a specific metric.
    """
    key: str
    type: MetricTypes
    per_container: Mapping[str, Measurement] = attr.Factory(dict)
    unit_hint: Optional[str] = None
    stats_filter: FrozenSet[str] = attr.Factory(frozenset)
    current_hook: Optional[Callable[['Metric'], Decimal]] = None


class MovingStatistics:
    __slots__ = (
        '_sum', '_count',
        '_min', '_max', '_last',
    )
    _sum: Decimal
    _count: int
    _min: Decimal
    _max: Decimal
    _last: List[Tuple[Decimal, float]]

    def __init__(self, initial_value: Decimal = None):
        self._last = []
        if initial_value is None:
            self._sum = Decimal(0)
            self._min = Decimal('inf')
            self._max = Decimal('-inf')
            self._count = 0
        else:
            self._sum = initial_value
            self._min = initial_value
            self._max = initial_value
            self._count = 1
            point = (initial_value, time.perf_counter())
            self._last.append(point)

    def update(self, value: Decimal):
        self._sum += value
        self._min = min(self._min, value)
        self._max = max(self._max, value)
        self._count += 1
        point = (value, time.perf_counter())
        self._last.append(point)
        # keep only the latest two data points
        if len(self._last) > 2:
            self._last.pop(0)

    @property
    def min(self) -> Decimal:
        return self._min

    @property
    def max(self) -> Decimal:
        return self._max

    @property
    def sum(self) -> Decimal:
        return self._sum

    @property
    def avg(self) -> Decimal:
        return self._sum / self._count

    @property
    def diff(self) -> Decimal:
        if len(self._last) == 2:
            return self._last[-1][0] - self._last[-2][0]
        return Decimal(0)

    @property
    def rate(self) -> Decimal:
        if len(self._last) == 2:
            return ((self._last[-1][0] - self._last[-2][0]) /
                    Decimal(self._last[-1][1] - self._last[-2][1]))
        return Decimal(0)

    def to_serializable_dict(self) -> MovingStatValue:
        q = Decimal('0.000')
        return {
            'min': str(remove_exponent(self.min.quantize(q))),
            'max': str(remove_exponent(self.max.quantize(q))),
            'sum': str(remove_exponent(self.sum.quantize(q))),
            'avg': str(remove_exponent(self.avg.quantize(q))),
            'diff': str(remove_exponent(self.diff.quantize(q))),
            'rate': str(remove_exponent(self.rate.quantize(q))),
            'version': 2,
        }


@attr.s(auto_attribs=True, slots=True)
class Metric:
    key: str
    type: MetricTypes
    stats: MovingStatistics
    stats_filter: FrozenSet[str]
    current: Decimal
    capacity: Optional[Decimal] = None
    unit_hint: Optional[str] = None
    current_hook: Optional[Callable[['Metric'], Decimal]] = None

    def update(self, value: Measurement):
        if value.capacity is not None:
            self.capacity = value.capacity
        self.stats.update(value.value)
        self.current = value.value
        if self.current_hook is not None:
            self.current = self.current_hook(self)

    def to_serializable_dict(self) -> MetricValue:
        q = Decimal('0.000')
        q_pct = Decimal('0.00')
        return {
            'current': str(remove_exponent(self.current.quantize(q))),
            'capacity': (str(remove_exponent(self.capacity.quantize(q)))
                         if self.capacity is not None else None),
            'pct': (
                str(remove_exponent(
                    (Decimal(self.current) / Decimal(self.capacity) * 100).quantize(q_pct)))
                if (self.capacity is not None and
                    self.capacity.is_normal() and
                    self.capacity > 0)
                else None),
            'unit_hint': self.unit_hint,
            **{f'stats.{k}': v  # type: ignore
               for k, v in self.stats.to_serializable_dict().items()
               if k in self.stats_filter},
        }


class StatContext:

    agent: 'AbstractAgent'
    mode: StatModes
    node_metrics: Mapping[MetricKey, Metric]
    device_metrics: Mapping[MetricKey, MutableMapping[DeviceId, Metric]]
    kernel_metrics: MutableMapping[KernelId, MutableMapping[MetricKey, Metric]]

    def __init__(self, agent: 'AbstractAgent', mode: StatModes = None, *,
                 cache_lifespan: int = 120) -> None:
        self.agent = agent
        self.mode = mode if mode is not None else StatModes.get_preferred_mode()
        self.cache_lifespan = cache_lifespan

        self.node_metrics = {}
        self.device_metrics = {}
        self.kernel_metrics = {}

        self._lock = asyncio.Lock()
        self._timestamps: MutableMapping[str, float] = {}

    def update_timestamp(self, timestamp_key: str) -> Tuple[float, float]:
        """
        Update the timestamp for the given key and return a pair of the current timestamp and
        the interval from the last update of the same key.

        If the last timestamp for the given key does not exist, the interval becomes "NaN".

        Intended to be used by compute plugins.
        """
        now = time.perf_counter()
        last = self._timestamps.get(timestamp_key, None)
        self._timestamps[timestamp_key] = now
        if last is None:
            return now, float('NaN')
        return now, now - last

    async def collect_node_stat(self):
        """
        Collect the per-node, per-device, and per-container statistics.

        Intended to be used by the agent.
        """
        async with self._lock:
            # Here we use asyncio.gather() instead of aiotools.TaskGroup
            # to keep methods of other plugins running when a plugin raises an error
            # instead of cancelling them.
            _tasks = []
            for computer in self.agent.computers.values():
                _tasks.append(computer.instance.gather_node_measures(self))
            results = await asyncio.gather(*_tasks, return_exceptions=True)
            for result in results:
                if isinstance(result, Exception):
                    log.error('collect_node_stat(): gather_node_measures() error',
                              exc_info=result)
                    continue
                for node_measure in result:
                    metric_key = node_measure.key
                    # update node metric
                    if metric_key not in self.node_metrics:
                        self.node_metrics[metric_key] = Metric(
                            metric_key, node_measure.type,
                            current=node_measure.per_node.value,
                            capacity=node_measure.per_node.capacity,
                            unit_hint=node_measure.unit_hint,
                            stats=MovingStatistics(node_measure.per_node.value),
                            stats_filter=frozenset(node_measure.stats_filter),
                            current_hook=node_measure.current_hook,
                        )
                    else:
                        self.node_metrics[metric_key].update(node_measure.per_node)
                    # update per-device metric
                    # NOTE: device IDs are defined by each metric keys.
                    for dev_id, measure in node_measure.per_device.items():
                        dev_id = str(dev_id)
                        if metric_key not in self.device_metrics:
                            self.device_metrics[metric_key] = {}
                        if dev_id not in self.device_metrics[metric_key]:
                            self.device_metrics[metric_key][dev_id] = Metric(
                                metric_key, node_measure.type,
                                current=measure.value,
                                capacity=measure.capacity,
                                unit_hint=node_measure.unit_hint,
                                stats=MovingStatistics(measure.value),
                                stats_filter=frozenset(node_measure.stats_filter),
                                current_hook=node_measure.current_hook,
                            )
                        else:
                            self.device_metrics[metric_key][dev_id].update(measure)

            # gather container metrics from compute plugins
            container_ids = []
            kernel_id_map = {}
            used_kernel_ids = set()
            for kid, kobj in self.agent.kernel_registry.items():
                cid = kobj['container_id']
                container_ids.append(cid)
                kernel_id_map[cid] = kid
                used_kernel_ids.add(kid)
            unused_kernel_ids = set(self.kernel_metrics.keys()) - used_kernel_ids
            for unused_kernel_id in unused_kernel_ids:
                log.debug('removing kernel_metric for {}', unused_kernel_id)
                self.kernel_metrics.pop(unused_kernel_id, None)
            _tasks = []
            for computer in self.agent.computers.values():
                _tasks.append(computer.instance.gather_container_measures(self, container_ids))
            for ctnr_measures in (await asyncio.gather(*_tasks, return_exceptions=True)):
                if isinstance(ctnr_measures, Exception):
                    log.error('gather_container_measures error',
                              exc_info=ctnr_measures)
                    continue
                for ctnr_measure in ctnr_measures:
                    metric_key = ctnr_measure.key
                    # update per-container metric
                    for cid, measure in ctnr_measure.per_container.items():
                        kernel_id = kernel_id_map[cid]
                        if kernel_id not in self.kernel_metrics:
                            self.kernel_metrics[kernel_id] = {}
                        if metric_key not in self.kernel_metrics[kernel_id]:
                            self.kernel_metrics[kernel_id][metric_key] = Metric(
                                metric_key, ctnr_measure.type,
                                current=measure.value,
                                capacity=measure.value,
                                unit_hint=ctnr_measure.unit_hint,
                                stats=MovingStatistics(measure.value),
                                stats_filter=frozenset(ctnr_measure.stats_filter),
                                current_hook=ctnr_measure.current_hook,
                            )
                        else:
                            self.kernel_metrics[kernel_id][metric_key].update(measure)

        # push to the Redis server
        redis_agent_updates = {
            'node': {
                key: obj.to_serializable_dict()
                for key, obj in self.node_metrics.items()
            },
            'devices': {
                metric_key: {dev_id: obj.to_serializable_dict()
                             for dev_id, obj in per_device.items()}
                for metric_key, per_device in self.device_metrics.items()
            },
        }
        if self.agent.local_config['debug']['log-stats']:
            log.debug('stats: node_updates: {0}: {1}',
                      self.agent.local_config['agent']['id'], redis_agent_updates['node'])
        serialized_agent_updates = msgpack.packb(redis_agent_updates)

        async def _pipe_builder(r: aioredis.Redis):
            async with r.pipeline() as pipe:
                pipe.set(self.agent.local_config['agent']['id'], serialized_agent_updates)
                pipe.expire(self.agent.local_config['agent']['id'], self.cache_lifespan)
                for kernel_id, metrics in self.kernel_metrics.items():
                    serialized_metrics = {
                        key: obj.to_serializable_dict()
                        for key, obj in metrics.items()
                    }
                    pipe.set(str(kernel_id), msgpack.packb(serialized_metrics))
                await pipe.execute()

        await redis.execute(self.agent.redis_stat_pool, _pipe_builder)

    async def collect_container_stat(
        self,
        container_ids: Sequence[ContainerId],
    ) -> Mapping[MetricKey, Metric]:
        """
        Collect the per-container statistics only,

        Intended to be used by the agent and triggered by container cgroup synchronization processes.
        """
        async with self._lock:
            kernel_id_map: Dict[ContainerId, KernelId] = {}
            for kid, info in self.agent.kernel_registry.items():
                cid = info['container_id']
                kernel_id_map[ContainerId(cid)] = kid
            # Here we use asyncio.gather() instead of aiotools.TaskGroup
            # to keep methods of other plugins running when a plugin raises an error
            # instead of cancelling them.
            _tasks = []
            kernel_id = None
            for computer in self.agent.computers.values():
                _tasks.append(asyncio.create_task(
                    computer.instance.gather_container_measures(self, container_ids),
                ))
            results = await asyncio.gather(*_tasks, return_exceptions=True)
            for result in results:
                if isinstance(result, Exception):
                    log.error('collect_container_stat(): gather_container_measures() error',
                              exc_info=result)
                    continue
                for ctnr_measure in result:
                    metric_key = ctnr_measure.key
                    # update per-container metric
                    for cid, measure in ctnr_measure.per_container.items():
                        try:
                            kernel_id = kernel_id_map[cid]
                        except KeyError:
                            continue
                        if kernel_id not in self.kernel_metrics:
                            self.kernel_metrics[kernel_id] = {}
                        if metric_key not in self.kernel_metrics[kernel_id]:
                            self.kernel_metrics[kernel_id][metric_key] = Metric(
                                metric_key, ctnr_measure.type,
                                current=measure.value,
                                capacity=measure.value,
                                unit_hint=ctnr_measure.unit_hint,
                                stats=MovingStatistics(measure.value),
                                stats_filter=frozenset(ctnr_measure.stats_filter),
                                current_hook=ctnr_measure.current_hook,
                            )
                        else:
                            self.kernel_metrics[kernel_id][metric_key].update(measure)

        if kernel_id is not None:
            metrics = self.kernel_metrics[kernel_id]
            serializable_metrics = {
                key: obj.to_serializable_dict()
                for key, obj in metrics.items()
            }
            if self.agent.local_config['debug']['log-stats']:
                log.debug('kernel_updates: {0}: {1}',
                          kernel_id, serializable_metrics)
            serialized_metrics = msgpack.packb(serializable_metrics)

            async def _pipe_builder(r: aioredis.Redis):
                async with r.pipeline() as pipe:
                    pipe.set(str(kernel_id), serialized_metrics)
                    await pipe.execute()

            await redis.execute(self.agent.redis_stat_pool, _pipe_builder)
            return metrics
        return {}



from __future__ import annotations

from abc import abstractmethod, ABCMeta
import asyncio
import codecs
from collections import OrderedDict, UserDict
from dataclasses import dataclass
import io
import json
import logging
import math
import re
import secrets
import time
from typing import (
    Any,
    Dict,
    FrozenSet,
    List,
    Literal,
    Mapping,
    Optional,
    Set,
    Sequence,
    Tuple,
    TypedDict,
    Union,
)

from async_timeout import timeout
import zmq, zmq.asyncio

from ai.backend.common import msgpack
from ai.backend.common.asyncio import current_loop
from ai.backend.common.docker import ImageRef
from ai.backend.common.enum_extension import StringSetFlag
from ai.backend.common.types import aobject, KernelId
from ai.backend.common.logging import BraceStyleAdapter
from .exception import UnsupportedBaseDistroError
from .resources import KernelResourceSpec

log = BraceStyleAdapter(logging.getLogger(__name__))

# msg types visible to the API client.
# (excluding control signals such as 'finished' and 'waiting-input'
# since they are passed as separate status field.)
ConsoleItemType = Literal[
    'stdout', 'stderr', 'media', 'html', 'log', 'completion',
]
outgoing_msg_types: FrozenSet[ConsoleItemType] = frozenset([
    'stdout', 'stderr', 'media', 'html', 'log', 'completion',
])
ResultType = Union[ConsoleItemType, Literal[
    'continued',
    'clean-finished',
    'build-finished',
    'finished',
    'exec-timeout',
    'waiting-input',
]]


class KernelFeatures(StringSetFlag):
    UID_MATCH = 'uid-match'
    USER_INPUT = 'user-input'
    BATCH_MODE = 'batch'
    QUERY_MODE = 'query'
    TTY_MODE = 'tty'


class ClientFeatures(StringSetFlag):
    INPUT = 'input'
    CONTINUATION = 'continuation'


# TODO: use Python 3.7 contextvars for per-client feature selection
default_client_features = frozenset({
    ClientFeatures.INPUT.value,
    ClientFeatures.CONTINUATION.value,
})
default_api_version = 4


class RunEvent(Exception):

    data: Any

    def __init__(self, data=None):
        super().__init__()
        self.data = data


class InputRequestPending(RunEvent):
    pass


class CleanFinished(RunEvent):
    pass


class BuildFinished(RunEvent):
    pass


class RunFinished(RunEvent):
    pass


class ExecTimeout(RunEvent):
    pass


@dataclass
class ResultRecord:
    msg_type: ResultType
    data: Optional[str] = None


class NextResult(TypedDict, total=False):
    runId: Optional[str]
    status: ResultType
    exitCode: Optional[int]
    options: Optional[Mapping[str, Any]]
    # v1
    stdout: Optional[str]
    stderr: Optional[str]
    media: Optional[Sequence[Any]]
    html: Optional[Sequence[Any]]
    # v2
    console: Optional[Sequence[Any]]


class AbstractKernel(UserDict, aobject, metaclass=ABCMeta):

    version: int
    agent_config: Mapping[str, Any]
    kernel_id: str
    image: ImageRef
    resource_spec: KernelResourceSpec
    service_ports: Any
    data: Dict[Any, Any]
    last_used: float
    termination_reason: Optional[str]
    clean_event: Optional[asyncio.Future]
    stats_enabled: bool

    _tasks: Set[asyncio.Task]

    runner: 'AbstractCodeRunner'

    def __init__(self, kernel_id: str, image: ImageRef, version: int, *,
                 agent_config: Mapping[str, Any],
                 resource_spec: KernelResourceSpec,
                 service_ports: Any,  # TODO: type-annotation
                 data: Dict[Any, Any]) -> None:
        self.agent_config = agent_config
        self.kernel_id = kernel_id
        self.image = image
        self.version = version
        self.resource_spec = resource_spec
        self.service_ports = service_ports
        self.data = data
        self.last_used = time.monotonic()
        self.termination_reason = None
        self.clean_event = None
        self.stats_enabled = False
        self._tasks = set()

    async def __ainit__(self) -> None:
        log.debug('kernel.__ainit__(k:{0}, api-ver:{1}, client-features:{2}): '
                  'starting new runner',
                  self.kernel_id, default_api_version, default_client_features)
        self.runner = await self.create_code_runner(
            client_features=default_client_features,
            api_version=default_api_version)

    def __getstate__(self) -> Mapping[str, Any]:
        props = self.__dict__.copy()
        del props['agent_config']
        del props['clean_event']
        del props['_tasks']
        return props

    def __setstate__(self, props) -> None:
        self.__dict__.update(props)
        # agent_config is set by the pickle.loads() caller.
        self.clean_event = None
        self._tasks = set()

    @abstractmethod
    async def close(self) -> None:
        """
        Release internal resources used for interacting with the kernel.
        Note that this does NOT terminate the container.
        """
        pass

    # We don't have "allocate_slots()" method here because:
    # - resource_spec is initialized by allocating slots at computer's alloc_map
    #   when creating new kernels.
    # - restoration from running containers is done by computer's classmethod
    #   "restore_from_container"

    def release_slots(self, computer_ctxs) -> None:
        """
        Release the resource slots occupied by the kernel
        to the allocation maps.
        """
        for accel_key, accel_alloc in self.resource_spec.allocations.items():
            computer_ctxs[accel_key].alloc_map.free(accel_alloc)

    @abstractmethod
    async def create_code_runner(
        self, *,
        client_features: FrozenSet[str],
        api_version: int,
    ) -> 'AbstractCodeRunner':
        raise NotImplementedError

    @abstractmethod
    async def check_status(self):
        raise NotImplementedError

    @abstractmethod
    async def get_completions(self, text, opts):
        raise NotImplementedError

    @abstractmethod
    async def get_logs(self):
        raise NotImplementedError

    @abstractmethod
    async def interrupt_kernel(self):
        raise NotImplementedError

    @abstractmethod
    async def start_service(self, service, opts):
        raise NotImplementedError

    @abstractmethod
    async def shutdown_service(self, service):
        raise NotImplementedError

    @abstractmethod
    async def get_service_apps(self):
        raise NotImplementedError

    @abstractmethod
    async def accept_file(self, filename, filedata):
        raise NotImplementedError

    @abstractmethod
    async def download_file(self, filepath):
        raise NotImplementedError

    @abstractmethod
    async def list_files(self, path: str):
        raise NotImplementedError

    async def execute(
        self,
        run_id: Optional[str],
        mode: Literal['batch', 'query', 'input', 'continue'],
        text: str,
        *,
        opts: Mapping[str, Any],
        api_version: int,
        flush_timeout: float,
    ) -> NextResult:
        myself = asyncio.current_task()
        assert myself is not None
        self._tasks.add(myself)
        try:
            await self.runner.attach_output_queue(run_id)
            try:
                if mode == 'batch':
                    await self.runner.feed_batch(opts)
                elif mode == 'query':
                    await self.runner.feed_code(text)
                elif mode == 'input':
                    await self.runner.feed_input(text)
                elif mode == 'continue':
                    pass
            except zmq.ZMQError:
                # cancel the operation by myself
                # since the peer is gone.
                raise asyncio.CancelledError
            return await self.runner.get_next_result(
                api_ver=api_version,
                flush_timeout=flush_timeout,
            )
        except asyncio.CancelledError:
            await self.runner.close()
            raise
        finally:
            self._tasks.remove(myself)


_zctx = None


class AbstractCodeRunner(aobject, metaclass=ABCMeta):

    kernel_id: KernelId
    started_at: float
    finished_at: Optional[float]
    exec_timeout: float
    max_record_size: int
    client_features: FrozenSet[str]

    input_sock: zmq.asyncio.Socket
    output_sock: zmq.asyncio.Socket

    completion_queue: asyncio.Queue[bytes]
    service_queue: asyncio.Queue[bytes]
    service_apps_info_queue: asyncio.Queue[bytes]
    status_queue: asyncio.Queue[bytes]
    output_queue: Optional[asyncio.Queue[ResultRecord]]
    current_run_id: Optional[str]
    pending_queues: OrderedDict[str, Tuple[asyncio.Event, asyncio.Queue[ResultRecord]]]

    read_task: Optional[asyncio.Task]
    status_task: Optional[asyncio.Task]
    watchdog_task: Optional[asyncio.Task]

    _closed: bool

    def __init__(
        self,
        kernel_id: KernelId,
        *,
        exec_timeout: float = 0,
        client_features: FrozenSet[str] = None,
    ) -> None:
        global _zctx
        self.kernel_id = kernel_id
        self.started_at = time.monotonic()
        self.finished_at = None
        if not math.isfinite(exec_timeout) or exec_timeout < 0:
            raise ValueError('execution timeout must be a zero or finite positive number.')
        self.kernel_id = kernel_id
        self.exec_timeout = exec_timeout
        self.max_record_size = 10 * (2 ** 20)  # 10 MBytes
        self.client_features = client_features or frozenset()
        if _zctx is None:
            _zctx = zmq.asyncio.Context()
        self.zctx = _zctx  # share the global context
        self.input_sock = self.zctx.socket(zmq.PUSH)
        self.output_sock = self.zctx.socket(zmq.PULL)
        self.completion_queue = asyncio.Queue(maxsize=128)
        self.service_queue = asyncio.Queue(maxsize=128)
        self.service_apps_info_queue = asyncio.Queue(maxsize=128)
        self.status_queue = asyncio.Queue(maxsize=128)
        self.output_queue = None
        self.pending_queues = OrderedDict()
        self.current_run_id = None
        self.read_task = None
        self.status_task = None
        self.watchdog_task = None
        self._closed = False

    async def __ainit__(self) -> None:
        loop = current_loop()
        self.input_sock.connect(await self.get_repl_in_addr())
        self.input_sock.setsockopt(zmq.LINGER, 50)
        self.output_sock.connect(await self.get_repl_out_addr())
        self.output_sock.setsockopt(zmq.LINGER, 50)
        self.status_task = loop.create_task(self.ping_status())
        self.read_task = loop.create_task(self.read_output())
        if self.exec_timeout > 0:
            self.watchdog_task = loop.create_task(self.watchdog())
        else:
            self.watchdog_task = None

    def __getstate__(self):
        props = self.__dict__.copy()
        del props['zctx']
        del props['input_sock']
        del props['output_sock']
        del props['completion_queue']
        del props['service_queue']
        del props['service_apps_info_queue']
        del props['status_queue']
        del props['output_queue']
        del props['pending_queues']
        del props['read_task']
        del props['status_task']
        del props['watchdog_task']
        del props['_closed']
        return props

    def __setstate__(self, props):
        global _zctx
        self.__dict__.update(props)
        if _zctx is None:
            _zctx = zmq.asyncio.Context()
        self.zctx = _zctx  # share the global context
        self.input_sock = self.zctx.socket(zmq.PUSH)
        self.output_sock = self.zctx.socket(zmq.PULL)
        self.completion_queue = asyncio.Queue(maxsize=128)
        self.service_queue = asyncio.Queue(maxsize=128)
        self.service_apps_info_queue = asyncio.Queue(maxsize=128)
        self.status_queue = asyncio.Queue(maxsize=128)
        self.output_queue = None
        self.pending_queues = OrderedDict()
        self.read_task = None
        self.status_task = None
        self.watchdog_task = None
        self._closed = False
        # __ainit__() is called by the caller.

    @abstractmethod
    async def get_repl_in_addr(self) -> str:
        raise NotImplementedError

    @abstractmethod
    async def get_repl_out_addr(self) -> str:
        raise NotImplementedError

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        try:
            if self.watchdog_task and not self.watchdog_task.done():
                self.watchdog_task.cancel()
                await self.watchdog_task
            if self.status_task and not self.status_task.done():
                self.status_task.cancel()
                await self.status_task
            if self.read_task and not self.read_task.done():
                self.read_task.cancel()
                await self.read_task
            if self.input_sock:
                self.input_sock.close()
            if self.output_sock:
                self.output_sock.close()
            # WARNING:
            # destroying zmq contexts here with possibility of re-entrance
            # may cause deadlocks.
        except Exception:
            log.exception("AbstractCodeRunner.close(): unexpected error")

    async def ping_status(self):
        """
        This is to keep the REPL in/out port mapping in the Linux
        kernel's NAT table alive.
        """
        try:
            while True:
                ret = await self.feed_and_get_status()
                if ret is None:
                    break
                await asyncio.sleep(10)
        except asyncio.CancelledError:
            pass
        except Exception:
            log.exception("AbstractCodeRunner.ping_status(): unexpected error")

    async def feed_batch(self, opts):
        if self.input_sock.closed:
            raise asyncio.CancelledError
        clean_cmd = opts.get('clean', '')
        if clean_cmd is None:
            clean_cmd = ''
        await self.input_sock.send_multipart([
            b'clean',
            clean_cmd.encode('utf8'),
        ])
        build_cmd = opts.get('build', '')
        if build_cmd is None:
            build_cmd = ''
        await self.input_sock.send_multipart([
            b'build',
            build_cmd.encode('utf8'),
        ])
        exec_cmd = opts.get('exec', '')
        if exec_cmd is None:
            exec_cmd = ''
        await self.input_sock.send_multipart([
            b'exec',
            exec_cmd.encode('utf8'),
        ])

    async def feed_code(self, text: str):
        if self.input_sock.closed:
            raise asyncio.CancelledError
        await self.input_sock.send_multipart([b'code', text.encode('utf8')])

    async def feed_input(self, text: str):
        if self.input_sock.closed:
            raise asyncio.CancelledError
        await self.input_sock.send_multipart([b'input', text.encode('utf8')])

    async def feed_interrupt(self):
        if self.input_sock.closed:
            raise asyncio.CancelledError
        await self.input_sock.send_multipart([b'interrupt', b''])

    async def feed_and_get_status(self):
        if self.input_sock.closed:
            raise asyncio.CancelledError
        await self.input_sock.send_multipart([b'status', b''])
        try:
            result = await self.status_queue.get()
            self.status_queue.task_done()
            return msgpack.unpackb(result)
        except asyncio.CancelledError:
            return None

    async def feed_and_get_completion(self, code_text, opts):
        if self.input_sock.closed:
            raise asyncio.CancelledError
        payload = {
            'code': code_text,
        }
        payload.update(opts)
        await self.input_sock.send_multipart([
            b'complete',
            json.dumps(payload).encode('utf8'),
        ])
        try:
            result = await self.completion_queue.get()
            self.completion_queue.task_done()
            return json.loads(result)
        except asyncio.CancelledError:
            return []

    async def feed_start_service(self, service_info):
        if self.input_sock.closed:
            raise asyncio.CancelledError
        await self.input_sock.send_multipart([
            b'start-service',
            json.dumps(service_info).encode('utf8'),
        ])
        try:
            with timeout(10):
                result = await self.service_queue.get()
            self.service_queue.task_done()
            return json.loads(result)
        except asyncio.CancelledError:
            return {'status': 'failed', 'error': 'cancelled'}
        except asyncio.TimeoutError:
            return {'status': 'failed', 'error': 'timeout'}

    async def feed_shutdown_service(self, service_name: str):
        if self.input_sock.closed:
            raise asyncio.CancelledError
        await self.input_sock.send_multipart([
            b'shutdown-service',
            json.dumps(service_name).encode('utf8'),
        ])

    async def feed_service_apps(self):
        await self.input_sock.send_multipart([
            b'get-apps',
            ''.encode('utf8'),
        ])
        try:
            with timeout(10):
                result = await self.service_apps_info_queue.get()
            self.service_apps_info_queue.task_done()
            return json.loads(result)
        except asyncio.CancelledError:
            return {'status': 'failed', 'error': 'cancelled'}
        except asyncio.TimeoutError:
            return {'status': 'failed', 'error': 'timeout'}

    async def watchdog(self) -> None:
        try:
            await asyncio.sleep(self.exec_timeout)
            if self.output_queue is not None:
                # TODO: what to do if None?
                self.output_queue.put(ResultRecord('exec-timeout', None))
        except asyncio.CancelledError:
            pass

    @staticmethod
    def aggregate_console(result: NextResult, records: Sequence[ResultRecord], api_ver: int) -> None:

        if api_ver == 1:

            stdout_items = []
            stderr_items = []
            media_items = []
            html_items = []

            for rec in records:
                if rec.msg_type == 'stdout':
                    stdout_items.append(rec.data or '')
                elif rec.msg_type == 'stderr':
                    stderr_items.append(rec.data or '')
                elif rec.msg_type == 'media' and rec.data is not None:
                    o = json.loads(rec.data)
                    media_items.append((o['type'], o['data']))
                elif rec.msg_type == 'html':
                    html_items.append(rec.data)

            result['stdout'] = ''.join(stdout_items)
            result['stderr'] = ''.join(stderr_items)
            result['media'] = media_items
            result['html'] = html_items

        elif api_ver >= 2:

            console_items: List[Tuple[ConsoleItemType, Union[str, Tuple[str, str]]]] = []
            last_stdout = io.StringIO()
            last_stderr = io.StringIO()

            for rec in records:

                if last_stdout.tell() and rec.msg_type != 'stdout':
                    console_items.append(('stdout', last_stdout.getvalue()))
                    last_stdout.seek(0)
                    last_stdout.truncate(0)
                if last_stderr.tell() and rec.msg_type != 'stderr':
                    console_items.append(('stderr', last_stderr.getvalue()))
                    last_stderr.seek(0)
                    last_stderr.truncate(0)

                if rec.msg_type == 'stdout':
                    last_stdout.write(rec.data or '')
                elif rec.msg_type == 'stderr':
                    last_stderr.write(rec.data or '')
                elif rec.msg_type == 'media' and rec.data is not None:
                    o = json.loads(rec.data)
                    console_items.append(('media', (o['type'], o['data'])))
                elif rec.msg_type in outgoing_msg_types:
                    # FIXME: currently mypy cannot handle dynamic specialization of literals.
                    console_items.append((rec.msg_type, rec.data))  # type: ignore

            if last_stdout.tell():
                console_items.append(('stdout', last_stdout.getvalue()))
            if last_stderr.tell():
                console_items.append(('stderr', last_stderr.getvalue()))

            result['console'] = console_items
            last_stdout.close()
            last_stderr.close()

        else:
            raise AssertionError('Unrecognized API version')

    async def get_next_result(self, api_ver=2, flush_timeout=2.0) -> NextResult:
        # Context: per API request
        has_continuation = ClientFeatures.CONTINUATION in self.client_features
        try:
            records = []
            result: NextResult
            assert self.output_queue is not None
            with timeout(flush_timeout if has_continuation else None):
                while True:
                    rec = await self.output_queue.get()
                    if rec.msg_type in outgoing_msg_types:
                        records.append(rec)
                    self.output_queue.task_done()
                    if rec.msg_type == 'finished':
                        data = json.loads(rec.data) if rec.data else {}
                        raise RunFinished(data)
                    elif rec.msg_type == 'clean-finished':
                        data = json.loads(rec.data) if rec.data else {}
                        raise CleanFinished(data)
                    elif rec.msg_type == 'build-finished':
                        data = json.loads(rec.data) if rec.data else {}
                        raise BuildFinished(data)
                    elif rec.msg_type == 'waiting-input':
                        opts = json.loads(rec.data) if rec.data else {}
                        raise InputRequestPending(opts)
                    elif rec.msg_type == 'exec-timeout':
                        raise ExecTimeout
        except asyncio.CancelledError:
            self.resume_output_queue()
            raise
        except asyncio.TimeoutError:
            result = {
                'runId': self.current_run_id,
                'status': 'continued',
                'exitCode': None,
                'options': None,
            }
            type(self).aggregate_console(result, records, api_ver)
            self.resume_output_queue()
            return result
        except CleanFinished as e:
            result = {
                'runId': self.current_run_id,
                'status': 'clean-finished',
                'exitCode': e.data.get('exitCode'),
                'options': None,
            }
            type(self).aggregate_console(result, records, api_ver)
            self.resume_output_queue()
            return result
        except BuildFinished as e:
            result = {
                'runId': self.current_run_id,
                'status': 'build-finished',
                'exitCode': e.data.get('exitCode'),
                'options': None,
            }
            type(self).aggregate_console(result, records, api_ver)
            self.resume_output_queue()
            return result
        except RunFinished as e:
            result = {
                'runId': self.current_run_id,
                'status': 'finished',
                'exitCode': e.data.get('exitCode'),
                'options': None,
            }
            type(self).aggregate_console(result, records, api_ver)
            self.next_output_queue()
            return result
        except ExecTimeout:
            result = {
                'runId': self.current_run_id,
                'status': 'exec-timeout',
                'exitCode': None,
                'options': None,
            }
            log.warning('Execution timeout detected on kernel '
                        f'{self.kernel_id}')
            type(self).aggregate_console(result, records, api_ver)
            self.next_output_queue()
            return result
        except InputRequestPending as e:
            result = {
                'runId': self.current_run_id,
                'status': 'waiting-input',
                'exitCode': None,
                'options': e.data,
            }
            type(self).aggregate_console(result, records, api_ver)
            self.resume_output_queue()
            return result
        except Exception:
            log.exception('unexpected error')
            raise

    async def attach_output_queue(self, run_id: Optional[str]) -> None:
        # Context: per API request
        if run_id is None:
            run_id = secrets.token_hex(16)
        assert run_id is not None
        if run_id not in self.pending_queues:
            q: asyncio.Queue[ResultRecord] = asyncio.Queue(maxsize=4096)
            activated = asyncio.Event()
            self.pending_queues[run_id] = (activated, q)
        else:
            activated, q = self.pending_queues[run_id]
        if self.output_queue is None:
            self.output_queue = q
        else:
            if self.current_run_id == run_id:
                # No need to wait if we are continuing.
                pass
            else:
                # If there is an outstanding ongoning execution,
                # wait until it has "finished".
                await activated.wait()
                activated.clear()
        self.current_run_id = run_id
        assert self.output_queue is q

    def resume_output_queue(self) -> None:
        """
        Use this to conclude get_next_result() when the execution should be
        continued from the client.

        At that time, we need to reuse the current run ID and its output queue.
        We don't change self.output_queue here so that we can continue to read
        outputs while the client sends the continuation request.
        """
        if self.current_run_id is None:
            return
        self.pending_queues.move_to_end(self.current_run_id, last=False)

    def next_output_queue(self) -> None:
        """
        Use this to conclude get_next_result() when we have finished a "run".
        """
        assert self.current_run_id is not None
        self.pending_queues.pop(self.current_run_id, None)
        self.current_run_id = None
        if len(self.pending_queues) > 0:
            # Make the next waiting API request handler to proceed.
            _, (activated, q) = self.pending_queues.popitem(last=False)
            self.output_queue = q
            activated.set()
        else:
            # If there is no pending request, just ignore all outputs
            # from the kernel.
            self.output_queue = None

    async def read_output(self) -> None:
        # We should use incremental decoder because some kernels may
        # send us incomplete UTF-8 byte sequences (e.g., Julia).
        decoders = (
            codecs.getincrementaldecoder('utf8')(errors='replace'),
            codecs.getincrementaldecoder('utf8')(errors='replace'),
        )
        while True:
            try:
                msg_type, msg_data = await self.output_sock.recv_multipart()
                try:
                    if msg_type == b'status':
                        await self.status_queue.put(msg_data)
                    elif msg_type == b'completion':
                        await self.completion_queue.put(msg_data)
                    elif msg_type == b'service-result':
                        await self.service_queue.put(msg_data)
                    elif msg_type == b'apps-result':
                        await self.service_apps_info_queue.put(msg_data)
                    elif msg_type == b'stdout':
                        if self.output_queue is None:
                            continue
                        if len(msg_data) > self.max_record_size:
                            msg_data = msg_data[:self.max_record_size]
                        await self.output_queue.put(
                            ResultRecord(
                                'stdout',
                                decoders[0].decode(msg_data),
                            ))
                    elif msg_type == b'stderr':
                        if self.output_queue is None:
                            continue
                        if len(msg_data) > self.max_record_size:
                            msg_data = msg_data[:self.max_record_size]
                        await self.output_queue.put(
                            ResultRecord(
                                'stderr',
                                decoders[1].decode(msg_data),
                            ))
                    else:
                        # Normal outputs should go to the current
                        # output queue.
                        if self.output_queue is None:
                            continue
                        await self.output_queue.put(
                            ResultRecord(
                                msg_type.decode('ascii'),
                                msg_data.decode('utf8'),
                            ))
                except asyncio.QueueFull:
                    pass
                if msg_type == b'build-finished':
                    # finalize incremental decoder
                    decoders[0].decode(b'', True)
                    decoders[1].decode(b'', True)
                elif msg_type == b'finished':
                    # finalize incremental decoder
                    decoders[0].decode(b'', True)
                    decoders[1].decode(b'', True)
                    self.finished_at = time.monotonic()
            except (asyncio.CancelledError, GeneratorExit):
                break
            except Exception:
                log.exception('unexpected error')
                break


def match_distro_data(data: Mapping[str, Any], distro: str) -> Tuple[str, Any]:
    """
    Find the latest or exactly matching entry from krunner_volumes mapping using the given distro
    string expression.

    It assumes that the keys of krunner_volumes mapping is a string concatenated with a distro
    prefix (e.g., "centos", "ubuntu") and a distro version composed of multiple integer components
    joined by single dots (e.g., "1.2.3", "18.04").
    """
    rx_ver_suffix = re.compile(r'(\d+(\.\d+)*)$')
    m = rx_ver_suffix.search(distro)
    if m is None:
        # Assume latest
        distro_prefix = distro
        distro_ver = None
    else:
        distro_prefix = distro[:-len(m.group(1))]
        distro_ver = m.group(1)

    # Check if there are static-build krunners first.
    if distro_prefix == 'alpine':
        libc_flavor = 'musl'
    else:
        libc_flavor = 'gnu'
    distro_key = f'static-{libc_flavor}'
    if volume := data.get(distro_key):
        return distro_key, volume

    # Search through the per-distro versions
    match_list = [
        (distro_key, value)
        for distro_key, value in data.items()
        if distro_key.startswith(distro_prefix)
    ]

    def _extract_version(item: Tuple[str, Any]) -> Tuple[int, ...]:
        m = rx_ver_suffix.search(item[0])
        if m is not None:
            return tuple(map(int, m.group(1).split('.')))
        return (0,)

    match_list = sorted(
        match_list,
        key=_extract_version,
        reverse=True)
    if match_list:
        if distro_ver is None:
            return match_list[0]
        for distro_key, value in match_list:
            if distro_key == distro:
                return (distro_key, value)
        return match_list[0]  # fallback to the latest of its kind
    raise UnsupportedBaseDistroError(distro)



import logging
import os
from pathlib import Path
import tempfile

from .. import BaseRunner

log = logging.getLogger()


class Runner(BaseRunner):

    log_prefix = 'lua-kernel'
    default_runtime_path = '/usr/local/bin/lua'
    default_child_env = {
        'TERM': 'xterm',
        'LANG': 'C.UTF-8',
        'SHELL': '/bin/ash',
        'USER': 'work',
        'HOME': '/home/work',
        'PATH': '/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin',
        'LD_LIBRARY_PATH': os.environ.get('LD_LIBRARY_PATH', ''),
        'LD_PRELOAD': os.environ.get('LD_PRELOAD', ''),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def init_with_loop(self):
        pass

    async def build_heuristic(self) -> int:
        log.info('no build process for lua language')
        return 0

    async def execute_heuristic(self) -> int:
        if Path('main.lua').is_file():
            cmd = [str(self.runtime_path), 'main.lua']
            return await self.run_subproc(cmd)
        else:
            log.error('cannot find executable ("main.lua").')
            return 127

    async def query(self, code_text) -> int:
        with tempfile.NamedTemporaryFile(suffix='.lua', dir='.') as tmpf:
            tmpf.write(code_text.encode('utf8'))
            tmpf.flush()
            cmd = [str(self.runtime_path), tmpf.name]
            return await self.run_subproc(cmd)

    async def complete(self, data):
        return []

    async def interrupt(self):
        # subproc interrupt is already handled by BaseRunner
        pass

    async def start_service(self, service_info):
        return None, {}



from subprocess import CalledProcessError
import asyncio


async def create_scratch_filesystem(scratch_dir, size):
    '''
    Create scratch folder size quota by using tmpfs filesystem.

    :param scratch_dir: The path of scratch directory.

    :param size: The quota size of scratch directory.
                 Size parameter is must be MiB(mebibyte).
    '''

    proc = await asyncio.create_subprocess_exec(*[
        'mount',
        '-t', 'tmpfs',
        '-o', f'size={size}M',
        'tmpfs', f'{scratch_dir}',
    ])
    exit_code = await proc.wait()

    if exit_code < 0:
        raise CalledProcessError(proc.returncode, proc.args,
                                 output=proc.stdout, stderr=proc.stderr)


async def destroy_scratch_filesystem(scratch_dir):
    '''
    Destroy scratch folder size quota by using tmpfs filesystem.

    :param scratch_dir: The path of scratch directory.
    '''
    proc = await asyncio.create_subprocess_exec(*[
        'umount',
        f'{scratch_dir}',
    ])
    exit_code = await proc.wait()

    if exit_code < 0:
        raise CalledProcessError(proc.returncode, proc.args,
                                 output=proc.stdout, stderr=proc.stderr)



import os

import trafaret as t

from ai.backend.common import config
from ai.backend.common import validators as tx

from .stats import StatModes
from .types import AgentBackend


coredump_defaults = {
    'enabled': False,
    'path': './coredumps',
    'backup-count': 10,
    'size-limit': '64M',
}

agent_local_config_iv = t.Dict({
    t.Key('agent'): t.Dict({
        tx.AliasedKey(['backend', 'mode']): tx.Enum(AgentBackend),
        t.Key('rpc-listen-addr', default=('', 6001)):
            tx.HostPortPair(allow_blank_host=True),
        t.Key('agent-sock-port', default=6007): t.Int[1024:65535],
        t.Key('id', default=None): t.Null | t.String,
        t.Key('ipc-base-path', default="/tmp/backend.ai/ipc"):
            tx.Path(type='dir', auto_create=True),
        t.Key('region', default=None): t.Null | t.String,
        t.Key('instance-type', default=None): t.Null | t.String,
        t.Key('scaling-group', default='default'): t.String,
        t.Key('pid-file', default=os.devnull):
            tx.Path(type='file', allow_nonexisting=True, allow_devnull=True),
        t.Key('event-loop', default='asyncio'): t.Enum('asyncio', 'uvloop'),
        t.Key('skip-manager-detection', default=False): t.ToBool,
        t.Key('aiomonitor-port', default=50102): t.Int[1:65535],
    }).allow_extra('*'),
    t.Key('container'): t.Dict({
        t.Key('kernel-uid', default=-1): tx.UserID,
        t.Key('kernel-gid', default=-1): tx.GroupID,
        t.Key('bind-host', default=''): t.String(allow_blank=True),
        t.Key('advertised-host', default=None): t.Null | t.String(),
        t.Key('port-range', default=(30000, 31000)): tx.PortRange,
        t.Key('stats-type', default='docker'):
            t.Null | t.Enum(*[e.value for e in StatModes]),
        t.Key('sandbox-type', default='docker'): t.Enum('docker', 'jail'),
        t.Key('jail-args', default=[]): t.List(t.String),
        t.Key('scratch-type'): t.Enum('hostdir', 'memory', 'k8s-nfs'),
        t.Key('scratch-root', default='./scratches'):
            tx.Path(type='dir', auto_create=True),
        t.Key('scratch-size', default='0'): tx.BinarySize,
        t.Key('scratch-nfs-address', default=None): t.Null | t.String,
        t.Key('scratch-nfs-options', default=None): t.Null | t.String,
    }).allow_extra('*'),
    t.Key('logging'): t.Any,  # checked in ai.backend.common.logging
    t.Key('resource'): t.Dict({
        t.Key('reserved-cpu', default=1): t.Int,
        t.Key('reserved-mem', default="1G"): tx.BinarySize,
        t.Key('reserved-disk', default="8G"): tx.BinarySize,
    }).allow_extra('*'),
    t.Key('debug'): t.Dict({
        t.Key('enabled', default=False): t.Bool,
        t.Key('skip-container-deletion', default=False): t.Bool,
        t.Key('log-stats', default=False): t.Bool,
        t.Key('log-kernel-config', default=False): t.Bool,
        t.Key('log-alloc-map', default=False): t.Bool,
        t.Key('log-events', default=False): t.Bool,
        t.Key('log-heartbeats', default=False): t.Bool,
        t.Key('log-docker-events', default=False): t.Bool,
        t.Key('coredump', default=coredump_defaults): t.Dict({
            t.Key('enabled', default=coredump_defaults['enabled']): t.Bool,
            t.Key('path', default=coredump_defaults['path']):
                tx.Path(type='dir', auto_create=True),
            t.Key('backup-count', default=coredump_defaults['backup-count']):
                t.Int[1:],
            t.Key('size-limit', default=coredump_defaults['size-limit']):
                tx.BinarySize,
        }).allow_extra('*'),
    }).allow_extra('*'),
}).merge(config.etcd_config_iv).allow_extra('*')

docker_extra_config_iv = t.Dict({
    t.Key('container'): t.Dict({
        t.Key('swarm-enabled', default=False): t.Bool,
    }).allow_extra('*'),
}).allow_extra('*')

default_container_logs_config = {
    'max-length': '10M',  # the maximum tail size
    'chunk-size': '64K',  # used when storing logs to Redis as a side-channel to the event bus
}

agent_etcd_config_iv = t.Dict({
    t.Key('container-logs', default=default_container_logs_config): t.Dict({
        t.Key('max-length', default=default_container_logs_config['max-length']): tx.BinarySize(),
        t.Key('chunk-size', default=default_container_logs_config['chunk-size']): tx.BinarySize(),
    }).allow_extra('*'),
}).allow_extra('*')

container_etcd_config_iv = t.Dict({
    t.Key('kernel-uid', optional=True): t.ToInt,
    t.Key('kernel-gid', optional=True): t.ToInt,
}).allow_extra('*')



import functools
from io import StringIO
from pathlib import Path
import uuid

import aiotools
from ai.backend.common.asyncio import current_loop
from ai.backend.common.docker import ImageRef
import asyncio
from decimal import Decimal
import hashlib
import logging
import pkg_resources
import random
import shutil
import signal
from typing import (
    Any,
    FrozenSet,
    List,
    Literal,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from ai.backend.common.logging_utils import BraceStyleAdapter

import cattr
from kubernetes_asyncio import (
    client as kube_client,
    config as kube_config,
)

from .resources import detect_resources
from .kernel import KubernetesKernel
from .kube_object import (
    ConfigMap, HostPathPersistentVolume, KubernetesAbstractVolume,
    KubernetesConfigMapVolume, KubernetesHostPathVolume,
    KubernetesPVCVolume, KubernetesVolumeMount, NFSPersistentVolume,
    PersistentVolumeClaim, Service,
)

from ..agent import ACTIVE_STATUS_SET, AbstractAgent, AbstractKernelCreationContext, ComputerContext
from ..exception import K8sError, UnsupportedResource
from ..kernel import KernelFeatures
from ..resources import (
    AbstractComputePlugin,
    KernelResourceSpec,
    Mount,
    known_slot_types,
)
from ..types import Container, ContainerStatus, Port

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
from ai.backend.common.types import (
    AutoPullBehavior,
    ClusterInfo,
    ContainerId,
    DeviceName,
    ImageRegistry,
    KernelCreationConfig,
    KernelId,
    MountPermission,
    MountTypes,
    ResourceSlot,
    SlotName,
    VFolderMount,
    current_resource_slots,
)

log = BraceStyleAdapter(logging.getLogger('ai.backend.agent.kubernetes.agent'))


class KubernetesKernelCreationContext(AbstractKernelCreationContext[KubernetesKernel]):
    config_map_name: str
    deployment_name: str
    scratch_dir: Path
    work_dir: Path
    config_dir: Path
    internal_mounts: List[Mount] = []
    static_pvc_name: str
    workers: Mapping[str, Mapping[str, str]]
    config_maps: List[ConfigMap]

    volume_mounts: List[KubernetesVolumeMount]
    volumes: List[KubernetesAbstractVolume]

    def __init__(
        self,
        kernel_id: KernelId,
        kernel_config: KernelCreationConfig,
        local_config: Mapping[str, Any],
        computers: MutableMapping[str, ComputerContext],
        workers: Mapping[str, Mapping[str, str]],
        static_pvc_name: str,
        restarting: bool = False,
    ) -> None:
        super().__init__(kernel_id, kernel_config, local_config, computers, restarting=restarting)
        scratch_dir = (self.local_config['container']['scratch-root'] / str(kernel_id)).resolve()

        self.scratch_dir = scratch_dir
        self.work_dir = scratch_dir / 'work'
        self.config_dir = scratch_dir / 'config'
        self.static_pvc_name = static_pvc_name
        self.workers = workers

        self.volume_mounts = []
        self.volumes = [
            KubernetesPVCVolume(
                name=f'kernel-{self.kernel_id}-scratches',
                persistentVolumeClaim={
                    'claimName': self.static_pvc_name,
                },
            ),
        ]

        self.config_maps = []

    async def get_extra_envs(self) -> Mapping[str, str]:
        return {}

    async def prepare_resource_spec(self) -> Tuple[KernelResourceSpec, Optional[Mapping[str, Any]]]:
        loop = current_loop()
        if self.restarting:
            await kube_config.load_kube_config()

            def _kernel_resource_spec_read():
                with open((self.config_dir / 'resource.txt').resolve(), 'r') as f:
                    resource_spec = KernelResourceSpec.read_from_file(f)
                return resource_spec

            resource_spec = await loop.run_in_executor(None, _kernel_resource_spec_read)
            resource_opts = None
        else:
            slots = ResourceSlot.from_json(self.kernel_config['resource_slots'])
            # Ensure that we have intrinsic slots.
            assert SlotName('cpu') in slots
            assert SlotName('mem') in slots
            # accept unknown slot type with zero values
            # but reject if they have non-zero values.
            for st, sv in slots.items():
                if st not in known_slot_types and sv != Decimal(0):
                    raise UnsupportedResource(st)
            # sanitize the slots
            current_resource_slots.set(known_slot_types)
            slots = slots.normalize_slots(ignore_unknown=True)
            resource_spec = KernelResourceSpec(
                container_id='',
                allocations={},
                slots={**slots},  # copy
                mounts=[],
                scratch_disk_size=0,  # TODO: implement (#70)
            )
            resource_opts = self.kernel_config.get('resource_opts', {})
        return resource_spec, resource_opts

    async def prepare_scratch(self) -> None:
        loop = current_loop()
        await kube_config.load_kube_config()
        core_api = kube_client.CoreV1Api()

        # Unlike Docker, static files will be mounted directly to blank folder
        # Check if NFS PVC for static files exists and bound
        nfs_pvc = await core_api.list_persistent_volume_claim_for_all_namespaces(
            label_selector='backend.ai/backend-ai-scratch-volume',
        )
        if len(nfs_pvc.items) == 0:
            raise K8sError('No PVC for backend.ai static files')
        pvc = nfs_pvc.items[0]
        if pvc.status.phase != 'Bound':
            raise K8sError('PVC not Bound')
        self.static_pvc_name = pvc.metadata.name

        def _create_scratch_dirs():
            self.work_dir.resolve().mkdir(parents=True, exist_ok=True)
            self.config_dir.resolve().mkdir(parents=True, exist_ok=True)

        # Mount scratch directory as PV
        # Config files can be mounted via ConfigMap
        await loop.run_in_executor(None, _create_scratch_dirs)

        if not self.restarting:
            # Since these files are bind-mounted inside a bind-mounted directory,
            # we need to touch them first to avoid their "ghost" files are created
            # as root in the host-side filesystem, which prevents deletion of scratch
            # directories when the agent is running as non-root.
            def _clone_dotfiles():
                jupyter_custom_css_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', 'jupyter-custom.css'))
                logo_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', 'logo.svg'))
                font_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', 'roboto.ttf'))
                font_italic_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', 'roboto-italic.ttf'))
                bashrc_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', '.bashrc'))
                bash_profile_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', '.bash_profile'))
                vimrc_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', '.vimrc'))
                tmux_conf_path = Path(pkg_resources.resource_filename(
                    'ai.backend.runner', '.tmux.conf'))
                jupyter_custom_dir = (self.work_dir / '.jupyter' / 'custom')
                jupyter_custom_dir.mkdir(parents=True, exist_ok=True)
                shutil.copy(jupyter_custom_css_path.resolve(), jupyter_custom_dir / 'custom.css')
                shutil.copy(logo_path.resolve(), jupyter_custom_dir / 'logo.svg')
                shutil.copy(font_path.resolve(), jupyter_custom_dir / 'roboto.ttf')
                shutil.copy(font_italic_path.resolve(), jupyter_custom_dir / 'roboto-italic.ttf')
                shutil.copy(bashrc_path.resolve(), self.work_dir / '.bashrc')
                shutil.copy(bash_profile_path.resolve(), self.work_dir / '.bash_profile')
                shutil.copy(vimrc_path.resolve(), self.work_dir / '.vimrc')
                shutil.copy(tmux_conf_path.resolve(), self.work_dir / '.tmux.conf')

            await loop.run_in_executor(None, _clone_dotfiles)

    async def get_intrinsic_mounts(self) -> Sequence[Mount]:
        mounts: List[Mount] = [
            # Mount scratch directory
            Mount(
                MountTypes.K8S_GENERIC,
                Path(str(self.kernel_id)),
                Path('/home/work'),
                MountPermission.READ_WRITE,
                opts={
                    'name': f'kernel-{self.kernel_id}-scratches',
                },
            ),
        ]

        # TODO: Find way to mount extra volumes

        return mounts

    async def apply_network(self, cluster_info: ClusterInfo) -> None:
        pass

    async def install_ssh_keypair(self, cluster_info: ClusterInfo) -> None:
        sshkey = cluster_info['ssh_keypair']
        if sshkey is None:
            return

        await kube_config.load_kube_config()
        core_api = kube_client.CoreV1Api()

        # Get hash of public key
        enc = hashlib.md5()
        enc.update(sshkey['public_key'].encode('ascii'))
        hash = enc.digest().decode('utf-8')

        try:
            await core_api.read_namespaced_config_map(f'ssh-keypair-{hash}', 'backend-ai')
        except:
            # Keypair not stored on ConfigMap, create one
            cm = ConfigMap('', f'kernel-{self.kernel_id}-ssh-keypair-{hash}')
            cm.put('public', sshkey['public_key'])
            cm.put('private', sshkey['private_key'])

            self.config_maps.append(cm)

        self.process_volumes([
            KubernetesConfigMapVolume(
                name=f'kernel-{self.kernel_id}-ssh-keypair',
                configMap={
                    'name': 'ssh-keypair-hash',
                },
            ),
        ])
        self.process_mounts([
            Mount(
                MountTypes.K8S_GENERIC,
                Path('public'),
                Path('/home/config/ssh/id_cluster.pub'),
                permission=MountPermission.READ_ONLY,
                opts={
                    'name': f'kernel-{self.kernel_id}-ssh-keypair',
                },
            ),
            Mount(
                MountTypes.K8S_GENERIC,
                Path('private'),
                Path('/home/config/ssh/id_cluster.pub'),
                permission=MountPermission.READ_ONLY,
                opts={
                    'name': f'kernel-{self.kernel_id}-ssh-keypair',
                },
            ),
        ])

    async def process_mounts(self, mounts: Sequence[Mount]):
        for i, mount in zip(range(len(mounts)), mounts):
            if mount.type == MountTypes.K8S_GENERIC:
                name = (mount.opts or {})['name']
                self.volume_mounts.append(
                    KubernetesVolumeMount(
                        name=name,
                        mountPath=mount.target.as_posix(),
                        subPath=mount.source.as_posix() if mount.source is not None else None,
                        readOnly=mount.permission == MountPermission.READ_ONLY,
                    ),
                )
            elif mount.type == MountTypes.K8S_HOSTPATH:
                name = (mount.opts or {})['name']
                self.volume_mounts.append(
                    KubernetesVolumeMount(
                        name=name,
                        mountPath=mount.target.as_posix(),
                        subPath=None,
                        readOnly=mount.permission == MountPermission.READ_ONLY,
                    ),
                )
            else:
                log.warn(
                    'Mount {}:{} -> Mount type {} it not supported on K8s Agent. Skipping mount',
                    mount.source,
                    mount.target,
                    mount.type,
                )

    def resolve_krunner_filepath(self, filename: str) -> Path:
        return Path(filename)

    def get_runner_mount(
        self,
        type: MountTypes,
        src: Union[str, Path],
        target: Union[str, Path],
        perm: Literal['ro', 'rw'] = 'ro',
        opts: Mapping[str, Any] = None,
    ) -> Mount:
        return Mount(
            MountTypes.K8S_GENERIC,
            Path(src),
            Path(target),
            MountPermission(perm),
            opts={
                **(opts or {}),
                'name': f'kernel-{self.kernel_id}-scratches',
            },
        )

    async def process_volumes(
        self,
        volumes: List[KubernetesAbstractVolume],
    ) -> None:
        self.volumes += volumes

    async def mount_vfolders(
        self,
        vfolders: Sequence[VFolderMount],
        resource_spec: KernelResourceSpec,
    ) -> None:
        # We can't mount vFolder backed by storage proxy
        for idx, vfolder in enumerate(vfolders):
            if self.internal_data.get('prevent_vfolder_mounts', False):
                # Only allow mount of ".logs" directory to prevent expose
                # internal-only information, such as Docker credentials to user's ".docker" vfolder
                # in image importer kernels.
                if vfolder.name != '.logs':
                    continue
            mount = Mount(
                MountTypes.K8S_HOSTPATH,
                Path(vfolder.host_path),
                Path(vfolder.kernel_path),
                vfolder.mount_perm,
                opts={
                    'name': f'kernel-{self.kernel_id}-hostPath-{idx}',
                },
            )
            self.process_volumes([
                KubernetesHostPathVolume(
                    name=f'kernel-{self.kernel_id}-hostPath-{idx}',
                    hostPath={
                        'path': vfolder.host_path.as_posix(),
                        'type': 'Directory',
                    },
                ),
            ])
            resource_spec.mounts.append(mount)

    async def apply_accelerator_allocation(self, computer, device_alloc) -> None:
        # update_nested_dict(
        #     self.computer_docker_args,
        #     await computer.generate_docker_args(self.docker, device_alloc))
        # TODO: add support for accelerator allocation
        pass

    async def generate_deployment_object(
        self, image: str, environ: Mapping[str, Any],
        ports: List[int], command: List[str],
        labels: Mapping[str, Any] = {},
    ) -> dict:
        return {
            'apiVersion': 'apps/v1',
            'kind': 'Deployment',
            'metadata': {
                'name': f'kernel-{self.kernel_id}',
                'labels': labels,
            },
            'spec': {
                'replicas': 0,
                'selector': {'matchLabels': {'run': f'kernel-{self.kernel_id}'}},
                'template': {
                    'metadata': {'labels': {'run': f'kernel-{self.kernel_id}'}},
                    'spec': {
                        'containers': [{
                            'name': f'kernel-{self.kernel_id}-session',
                            'image': image,
                            'imagePullPolicy': 'IfNotPresent',
                            'command': ['sh', '/opt/kernel/entrypoint.sh'],
                            'args': command,
                            'env': [{'name': k, 'value': v} for k, v in environ.items()],
                            'volumeMounts': [cattr.unstructure(v) for v in self.volume_mounts],
                            'ports': [{'containerPort': x} for x in ports],
                        }],
                        'volumes': [cattr.unstructure(v) for v in self.volumes],
                    },
                },
            },
        }

    async def spawn(
        self,
        resource_spec: KernelResourceSpec,
        resource_opts,
        environ: Mapping[str, str],
        service_ports,
        preopen_ports,
        cmdargs: List[str],
    ) -> KubernetesKernel:
        await kube_config.load_kube_config()
        core_api = kube_client.CoreV1Api()
        apps_api = kube_client.AppsV1Api()

        loop = current_loop()
        image_labels = self.kernel_config['image']['labels']
        exposed_ports = [2000, 2001]
        for sport in service_ports:
            exposed_ports.extend(sport['container_ports'])

        if self.restarting:
            pass
        else:
            if bootstrap := self.kernel_config.get('bootstrap_script'):

                def _write_user_bootstrap_script():
                    (self.work_dir / 'bootstrap.sh').write_text(bootstrap)
                    if KernelFeatures.UID_MATCH in self.kernel_features:  # UID Match won't work on K8s
                        # uid = self.local_config['container']['kernel-uid']
                        # gid = self.local_config['container']['kernel-gid']
                        # if os.geteuid() == 0:
                        #     os.chown(self.work_dir / 'bootstrap.sh', uid, gid)
                        pass

                await loop.run_in_executor(None, _write_user_bootstrap_script)

            def _write_config(file_name: str, content: str):
                (self.config_dir / file_name).write_text(content)

            with StringIO() as buf:
                for k, v in environ.items():
                    buf.write(f'{k}={v}\n')
                # accel_envs = self.computer_docker_args.get('Env', [])
                # for env in accel_envs:
                #     buf.write(f'{env}\n')
                await loop.run_in_executor(
                    None, functools.partial(_write_config, 'environ.txt', buf.getvalue()),
                )

            with StringIO() as buf:
                resource_spec.write_to_file(buf)
                for dev_type, device_alloc in resource_spec.allocations.items():
                    computer_self = self.computers[dev_type]
                    kvpairs = \
                        await computer_self.instance.generate_resource_data(device_alloc)
                    for k, v in kvpairs.items():
                        buf.write(f'{k}={v}\n')
                await loop.run_in_executor(
                    None, functools.partial(_write_config, 'resource.txt', buf.getvalue()),
                )

            docker_creds = self.internal_data.get('docker_credentials')
            if docker_creds:
                await loop.run_in_executor(
                    None, functools.partial(_write_config, 'docker-creds.json', docker_creds),
                )

        if keypair := self.internal_data.get('ssh_keypair'):
            for mount in resource_spec.mounts:
                container_path = str(mount).split(':')[1]
                if container_path == '/home/work/.ssh':
                    break
            else:
                pubkey = keypair['public_key'].encode('ascii')
                privkey = keypair['private_key'].encode('ascii')
                ssh_config_map = ConfigMap(self.kernel_id, f'kernel-{self.kernel_id}-ssh-config')
                ssh_config_map.put('authorized_keys', pubkey)
                ssh_config_map.put('id_container', privkey)
                self.process_volumes([
                    KubernetesConfigMapVolume(
                        name='ssh-config',
                        configMap={
                            'name': f'kernel-{self.kernel_id}-ssh-config',
                        },
                    ),
                ])
                self.process_mounts([
                    Mount(
                        MountTypes.K8S_GENERIC,
                        Path('authorized_keys'),
                        Path('/home/work/.ssh/authorized_keys'),
                        opts={
                            'name': 'ssh-config',
                        },
                    ),
                    Mount(
                        MountTypes.K8S_GENERIC,
                        Path('id_container'),
                        Path('/home/work/.ssh/id_container'),
                        opts={
                            'name': 'ssh-config',
                        },
                    ),
                ])

        # higher priority dotfiles are stored last to support overwriting
        for dotfile in self.internal_data.get('dotfiles', []):
            if dotfile['path'].startswith('/'):
                if dotfile['path'].startswith('/home/'):
                    path_arr = dotfile['path'].split('/')
                    file_path: Path = self.scratch_dir / '/'.join(path_arr[2:])
                else:
                    file_path = Path(dotfile['path'])
            else:
                file_path = self.work_dir / dotfile['path']
            file_path.parent.mkdir(parents=True, exist_ok=True)
            await loop.run_in_executor(
                None,
                file_path.write_text,
                dotfile['data'])

        # TODO: Mark shmem feature as unsupported when advertising agent

        # PHASE 4: Run!

        encoded_preopen_ports = ','.join(f'{port_no}:preopen:{port_no}' for port_no in preopen_ports)

        service_port_label = image_labels['ai.backend.service-ports']
        if len(encoded_preopen_ports) > 0:
            service_port_label += f',{encoded_preopen_ports}'

        deployment = await self.generate_deployment_object(
            self.image_ref.canonical,
            environ,
            exposed_ports,
            cmdargs,
            labels={
                'ai.backend.service-ports': service_port_label.replace(':', '-').replace(',', '.'),
            },
        )

        if self.local_config['debug']['log-kernel-config']:
            log.debug('Initial container config: {0}', deployment)

        expose_service = Service(
            str(self.kernel_id),
            f'kernel-{self.kernel_id}-expose',
            [
                (port, f'kernel-{self.kernel_id}-svc-{index}')
                for index, port in zip(range(len(exposed_ports)), exposed_ports)
            ],
        )

        async def rollup(
            functions: List[Tuple[Optional[functools.partial],
            Optional[functools.partial]]],
        ):
            rollback_functions: List[Optional[functools.partial]] = []

            for (rollup_function, future_rollback_function) in functions:
                try:
                    if rollup_function:
                        await rollup_function()
                    rollback_functions.append(future_rollback_function)
                except Exception as e:
                    for rollback_function in rollback_functions[::-1]:
                        if rollback_function:
                            await rollback_function()
                    log.exception('Error while rollup: {}', e)
                    raise

        arguments: List[Tuple[Optional[functools.partial], Optional[functools.partial]]] = []
        node_ports = []

        try:
            expose_service_api_response = \
                await core_api.create_namespaced_service('backend-ai', body=expose_service.to_dict())
        except Exception as e:
            log.exception('Error while rollup: {}', e)
            raise

        node_ports = expose_service_api_response.spec.ports
        arguments.append((
            None,
            functools.partial(core_api.delete_namespaced_service, expose_service.name, 'backend-ai'),
        ))
        for cm in self.config_maps:
            arguments.append((
                functools.partial(
                    core_api.create_namespaced_config_map, 'backend-ai',
                    body=cm.to_dict(),
                ),
                functools.partial(core_api.delete_namespaced_config_map, cm.name, 'backend-ai'),
            ))

        arguments.append((
            functools.partial(
                apps_api.create_namespaced_deployment, 'backend-ai',
                body=deployment, pretty='pretty-example',
            ),
            None,
        ))

        await rollup(arguments)

        assigned_ports: MutableMapping[str, int] = {}
        for port in node_ports:
            assigned_ports[port.port] = port.node_port

        ctnr_host_port_map: MutableMapping[int, int] = {}
        stdin_port = 0
        stdout_port = 0
        for idx, port in enumerate(exposed_ports):
            host_port = assigned_ports[port]

            if port == 2000:     # intrinsic
                repl_in_port = host_port
            elif port == 2001:   # intrinsic
                repl_out_port = host_port
            elif port == 2002:   # legacy
                stdin_port = host_port
            elif port == 2003:   # legacy
                stdout_port = host_port
            else:
                ctnr_host_port_map[port] = host_port
        for sport in service_ports:
            sport['host_ports'] = tuple(
                ctnr_host_port_map[cport] for cport in sport['container_ports']
            )

        target_node_ip = random.choice([x['InternalIP'] for x in self.workers.values()])

        kernel_obj = await KubernetesKernel.new(
            self.kernel_id,
            self.image_ref,
            self.kspec_version,
            agent_config=self.local_config,
            service_ports=service_ports,
            resource_spec=resource_spec,
            data={
                'container_id': '',
                'kernel_host': target_node_ip,
                'repl_in_port': repl_in_port,
                'repl_out_port': repl_out_port,
                'stdin_port': stdin_port,    # legacy
                'stdout_port': stdout_port,  # legacy
                'assigned_ports': assigned_ports,
                # 'domain_socket_proxies': self.domain_socket_proxies,
                'block_service_ports': self.internal_data.get('block_service_ports', False),
            })
        return kernel_obj


class KubernetesAgent(AbstractAgent[KubernetesKernel, KubernetesKernelCreationContext]):
    workers: MutableMapping[str, MutableMapping[str, str]] = {}
    k8s_ptask_group: aiotools.PersistentTaskGroup

    def __init__(
        self,
        etcd: AsyncEtcd,
        local_config: Mapping[str, Any],
        *,
        stats_monitor: StatsPluginContext,
        error_monitor: ErrorPluginContext,
        skip_initial_scan: bool = False,
    ) -> None:
        super().__init__(
            etcd,
            local_config,
            stats_monitor=stats_monitor,
            error_monitor=error_monitor,
            skip_initial_scan=skip_initial_scan,
        )

    async def __ainit__(self) -> None:
        await super().__ainit__()
        ipc_base_path = self.local_config['agent']['ipc-base-path']
        self.agent_sockpath = ipc_base_path / 'container' / f'agent.{self.local_instance_id}.sock'

        await self.check_krunner_pv_status()
        await self.fetch_workers()
        self.k8s_ptask_group = aiotools.PersistentTaskGroup()
        # Socket Relay initialization
        # Agent socket handler initialization
        # K8s event monitor task initialization

    async def check_krunner_pv_status(self):
        capacity = format(self.local_config['container']['scratch-size'], 'g')[:-1]

        await kube_config.load_kube_config()
        core_api = kube_client.CoreV1Api()

        namespaces = await core_api.list_namespace()
        if len(list(filter(lambda ns: ns.metadata.name == 'backend-ai', namespaces.items))) == 0:
            await core_api.create_namespace({
                'apiVersion': 'v1',
                'kind': 'Namespace',
                'metadata': {
                    'name': 'backend-ai',
                },
            })

        pv = await core_api.list_persistent_volume(label_selector='backend.ai/backend-ai-scratch-volume')

        if len(pv.items) == 0:
            # PV does not exists; create one
            if self.local_config['container']['scratch-type'] == 'k8s-nfs':
                new_pv = NFSPersistentVolume(
                    self.local_config['container']['scratch-nfs-address'],
                    'backend-ai-static-pv', capacity,
                )
                new_pv.label(
                    'backend.ai/backend-ai-scratch-volume',
                    self.local_config['container']['scratch-nfs-address'],
                )
                new_pv.options = [
                    x.strip()
                    for x in self.local_config['container']['scratch-nfs-options'].split(',')
                ]
            elif self.local_config['container']['scratch-type'] == 'hostdir':
                new_pv = HostPathPersistentVolume(
                    self.local_config['container']['scratch-root'].as_posix(),
                    'backend-ai-static-pv', capacity,
                )
                new_pv.label('backend.ai/backend-ai-scratch-volume', 'hostPath')
            else:
                raise NotImplementedError(
                    f'Scratch type {self.local_config["container"]["scratch-type"]} is not supported',
                )

            try:
                await core_api.create_persistent_volume(body=new_pv.to_dict())
            except:
                raise

        pvc = await core_api.list_persistent_volume_claim_for_all_namespaces(
            label_selector='backend.ai/backend-ai-scratch-volume',
        )

        if len(pvc.items) == 0:
            # PV does not exists; create one
            new_pvc = PersistentVolumeClaim(
                'backend-ai-static-pvc', 'backend-ai-static-pv', capacity,
            )
            if self.local_config['container']['scratch-type'] == 'k8s-nfs':
                new_pvc.label(
                    'backend.ai/backend-ai-scratch-volume',
                    self.local_config['container']['scratch-nfs-address'],
                )
            else:
                new_pvc.label('backend.ai/backend-ai-scratch-volume', 'hostPath')
            try:
                await core_api.create_namespaced_persistent_volume_claim(
                    'backend-ai', body=new_pvc.to_dict(),
                )
            except Exception as e:
                log.exception('Error: {}', e)
                raise

    async def fetch_workers(self):
        await kube_config.load_kube_config()
        core_api = kube_client.CoreV1Api()
        nodes = await core_api.list_node()
        for node in nodes.items:
            # if 'node-role.kubernetes.io/master' in node.metadata.labels.keys():
            #     continue
            self.workers[node.metadata.name] = node.status.capacity
            for addr in node.status.addresses:
                if addr.type == 'InternalIP':
                    self.workers[node.metadata.name]['InternalIP'] = addr.address
                if addr.type == 'ExternalIP':
                    self.workers[node.metadata.name]['ExternalIP'] = addr.address

    async def shutdown(self, stop_signal: signal.Signals):
        # Stop agent socket handler task

        try:
            if self.k8s_ptask_group is not None:
                await self.k8s_ptask_group.shutdown()
            await super().shutdown(stop_signal)
        finally:
            # Stop k8s event monitoring.
            pass

    async def detect_resources(self) -> Tuple[
        Mapping[DeviceName, AbstractComputePlugin],
        Mapping[SlotName, Decimal],
    ]:
        return await detect_resources(self.etcd, self.local_config)

    async def enumerate_containers(
        self,
        status_filter: FrozenSet[ContainerStatus] = ACTIVE_STATUS_SET,
    ) -> Sequence[Tuple[KernelId, Container]]:
        await kube_config.load_kube_config()
        core_api = kube_client.CoreV1Api()

        result = []
        fetch_tasks = []
        for deployment in (await core_api.list_namespaced_pod('backend-ai')).items:
            # Additional check to filter out real worker pods only?

            async def _fetch_container_info(pod: Any):
                kernel_id: Union[str, None] = "(unknown)"
                try:
                    kernel_id = await get_kernel_id_from_deployment(pod)
                    if kernel_id is None:
                        return
                    # Is it okay to assume that only one container resides per pod?
                    if pod['status']['containerStatuses'][0]['stats'].keys()[0] in status_filter:
                        result.append(
                            (
                                KernelId(uuid.UUID(kernel_id)),
                                await container_from_pod(pod),
                            ),
                        )
                except asyncio.CancelledError:
                    pass
                except Exception:
                    log.exception(
                        "error while fetching container information (cid:{}, k:{})",
                        pod['metadata']['uid'], kernel_id,
                    )

            fetch_tasks.append(_fetch_container_info(deployment))

        await asyncio.gather(*fetch_tasks, return_exceptions=True)
        return result

    async def scan_images(self) -> Mapping[str, str]:
        # Retrieving image label from registry api is not possible
        return {}

    async def handle_agent_socket(self):
        # TODO: Add support for remote agent socket mechanism
        pass

    async def pull_image(self, image_ref: ImageRef, registry_conf: ImageRegistry) -> None:
        # TODO: Add support for appropriate image pulling mechanism on K8s
        pass

    async def check_image(self, image_ref: ImageRef, image_id: str, auto_pull: AutoPullBehavior) -> bool:
        # TODO: Add support for appropriate image checking mechanism on K8s
        # Just mark all images as 'pulled' since we can't manually initiate image pull on each kube node
        return True

    async def init_kernel_context(
        self,
        kernel_id: KernelId,
        kernel_config: KernelCreationConfig,
        *,
        restarting: bool = False,
    ) -> KubernetesKernelCreationContext:
        return KubernetesKernelCreationContext(
            kernel_id,
            kernel_config,
            self.local_config,
            self.computers,
            self.workers,
            'backend-ai-static-pvc',
            restarting=restarting,
        )

    async def destroy_kernel(self, kernel_id: KernelId, container_id: Optional[ContainerId]) -> None:
        await kube_config.load_kube_config()
        core_api = kube_client.CoreV1Api()
        apps_api = kube_client.AppsV1Api()

        async def force_cleanup(reason='self-terminated'):
            await self.send_event('kernel_terminated',
                                  kernel_id, 'self-terminated',
                                  None)
        try:
            kernel = self.kernel_registry[kernel_id]
        except:
            log.warning('_destroy_kernel({0}) kernel missing (already dead?)',
                        kernel_id)
            await asyncio.shield(self.k8s_ptask_group.create_task(force_cleanup()))
            return None
        deployment_name = kernel['deployment_name']
        try:
            await core_api.delete_namespaced_service(f'{deployment_name}-service', 'backend-ai')
            await core_api.delete_namespaced_service(f'{deployment_name}-nodeport', 'backend-ai')
            await apps_api.delete_namespaced_deployment(f'{deployment_name}', 'backend-ai')
        except:
            log.warning('_destroy({0}) kernel missing (already dead?)', kernel_id)

    async def clean_kernel(
        self,
        kernel_id: KernelId,
        container_id: Optional[ContainerId],
        restarting: bool,
    ) -> None:
        loop = current_loop()
        if not restarting:
            scratch_dir = self.local_config['container']['scratch-root'] / str(kernel_id)
            await loop.run_in_executor(None, shutil.rmtree, str(scratch_dir))

    async def create_overlay_network(self, network_name: str) -> None:
        return await super().create_overlay_network(network_name)

    async def destroy_overlay_network(self, network_name: str) -> None:
        return await super().destroy_overlay_network(network_name)

    async def create_local_network(self, network_name: str) -> None:
        return await super().create_local_network(network_name)

    async def destroy_local_network(self, network_name: str) -> None:
        return await super().destroy_local_network(network_name)

    async def restart_kernel__load_config(
        self,
        kernel_id: KernelId,
        name: str,
    ) -> bytes:
        loop = current_loop()
        scratch_dir = (self.local_config['container']['scratch-root'] / str(kernel_id)).resolve()
        config_dir = scratch_dir / 'config'
        return await loop.run_in_executor(
            None,
            (config_dir / name).read_bytes,
        )

    async def restart_kernel__store_config(
        self,
        kernel_id: KernelId,
        name: str,
        data: bytes,
    ) -> None:
        loop = current_loop()
        scratch_dir = (self.local_config['container']['scratch-root'] / str(kernel_id)).resolve()
        config_dir = scratch_dir / 'config'
        return await loop.run_in_executor(
            None,
            (config_dir / name).write_bytes,
            data,
        )


async def get_kernel_id_from_deployment(pod: Any) -> Optional[str]:
    # TODO: create function which extracts kernel id from pod object
    return pod.get('metadata', {}).get('name')


async def container_from_pod(pod: Any) -> Container:
    status: ContainerStatus = ContainerStatus.RUNNING
    phase = pod['status']['phase']
    if phase == 'Pending' or phase == 'Running':
        status = ContainerStatus.RUNNING
    elif phase == 'Succeeded':
        status = ContainerStatus.EXITED
    elif phase == 'Failed' or phase == 'Unknown':
        status = ContainerStatus.DEAD

    # TODO: Create Container object from K8s Pod definition
    return Container(
        id=ContainerId(''),
        status=status,
        image=pod['spec']['containers'][0]['image'],
        labels=pod['metadata']['labels'],
        ports=[Port(
            port['host_ip'], port['container_port'], port['host_port'],
        ) for port in pod['spec']['containers'][0]['ports']],
        backend_obj=pod,
    )



import asyncio
import enum
from typing import (
    Any, Optional,
    Mapping,
    Sequence,
)

import attr

from ai.backend.common.types import (
    ContainerId,
    KernelId,
)


class AgentBackend(enum.Enum):
    # The list of importable backend names under "ai.backend.agent" pkg namespace.
    DOCKER = 'docker'
    KUBERNETES = 'kubernetes'


@attr.s(auto_attribs=True, slots=True)
class VolumeInfo:
    name: str             # volume name
    container_path: str   # in-container path as str
    mode: str             # 'rw', 'ro', 'rwm'


@attr.s(auto_attribs=True, slots=True)
class Port:
    host: str
    private_port: int
    host_port: int


class ContainerStatus(str, enum.Enum):
    RUNNING = 'running'
    RESTARTING = 'restarting'
    PAUSED = 'paused'
    EXITED = 'exited'
    DEAD = 'dead'
    REMOVING = 'removing'


@attr.s(auto_attribs=True, slots=True)
class Container:
    id: ContainerId
    status: ContainerStatus
    image: str
    labels: Mapping[str, str]
    ports: Sequence[Port]
    backend_obj: Any  # used to keep the backend-specific data


class LifecycleEvent(int, enum.Enum):
    DESTROY = 0
    CLEAN = 1
    START = 2


@attr.s(auto_attribs=True, slots=True)
class ContainerLifecycleEvent:
    kernel_id: KernelId
    container_id: Optional[ContainerId]
    event: LifecycleEvent
    reason: str
    done_future: Optional[asyncio.Future] = None
    exit_code: Optional[int] = None
    suppress_events: bool = False

    def __str__(self):
        if self.container_id:
            cid = self.container_id[:13]
        else:
            cid = 'unknown'
        return (
            f"LifecycleEvent("
            f"{self.event.name}, "
            f"k:{self.kernel_id}, "
            f"c:{cid}, "
            f"reason:{self.reason!r})"
        )



import asyncio
from pathlib import Path

from async_timeout import timeout

__all__ = (
    'current_loop',
    'find_executable',
    'safe_close_task',
    'wait_local_port_open',
)


if hasattr(asyncio, 'get_running_loop'):
    current_loop = asyncio.get_running_loop  # type: ignore  # noqa
else:
    current_loop = asyncio.get_event_loop    # type: ignore  # noqa


def find_executable(*paths):
    '''
    Find the first executable regular file in the given list of paths.
    '''
    for path in paths:
        if isinstance(path, (str, bytes)):
            path = Path(path)
        if not path.exists():
            continue
        for child in path.iterdir():
            if child.is_file() and child.stat().st_mode & 0o100 != 0:
                return child
    return None


async def safe_close_task(task):
    if task is not None and not task.done():
        task.cancel()
        await task


async def wait_local_port_open(port):
    while True:
        try:
            with timeout(10.0):
                reader, writer = await asyncio.open_connection('127.0.0.1', port)
        except ConnectionRefusedError:
            await asyncio.sleep(0.1)
            continue
        except asyncio.TimeoutError:
            raise
        except Exception:
            raise
        else:
            writer.close()
            if hasattr(writer, 'wait_closed'):
                await writer.wait_closed()
            break



'''
The kernel main program.
'''

import argparse
import importlib
import os
from pathlib import Path
import signal
import sys

import uvloop

from . import lang_map
from .compat import asyncio_run_forever


def parse_args(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', action='store_true', default=False)
    parser.add_argument('lang', type=str, choices=lang_map.keys())
    parser.add_argument('runtime_path', type=Path, nargs='?', default=None)
    return parser.parse_args(args)


def main(args) -> None:
    cls_name = lang_map[args.lang]
    imp_path, cls_name = cls_name.rsplit('.', 1)
    mod = importlib.import_module(imp_path)
    cls = getattr(mod, cls_name)

    if args.runtime_path is None:
        runtime_path = cls.default_runtime_path
    else:
        runtime_path = args.runtime_path
    runner = cls(runtime_path)

    # Replace stdin with a "null" file
    # (trying to read stdin will raise EOFError immediately afterwards.)
    sys.stdin = open(os.devnull, 'r', encoding='latin1')
    asyncio_run_forever(
        runner._init(args),
        runner._shutdown(),
        stop_signals={signal.SIGINT, signal.SIGTERM},
    )


args = parse_args()
uvloop.install()
main(args)
